# !/usr/bin/python
# -*- coding: latin-1 -*-
"""
A module for creating and managing grids of model spectra
"""
from functools import partial
from glob import glob
import multiprocessing
import os
import pickle
from pkg_resources import resource_filename
import time
import warnings

from astropy.io import fits
from astropy.utils.exceptions import AstropyWarning
import astropy.table as at
import astropy.units as q
import h5py
import numpy as np
from scipy.interpolate import RegularGridInterpolator
from scipy.ndimage import zoom

from . import utils

warnings.simplefilter('ignore', category=AstropyWarning)
warnings.simplefilter('ignore', category=FutureWarning)


class ModelGrid(object):
    """
    Creates a ModelGrid object which contains a multi-parameter
    grid of model spectra and its references

    Attributes
    ----------
    path: str
        The path to the directory of FITS files used to create the ModelGrid
    refs: list, str
        The references for the data contained in the ModelGrid
    teff_rng: tuple
        The range of effective temperatures [K]
    logg_rng: tuple
        The range of surface gravities [dex]
    FeH_rng: tuple
        The range of metalicities [dex]
    wave_rng: array-like
        The wavelength range of the models [um]
    n_bins: int
        The number of bins for the ModelGrid wavelength array
    data: astropy.table.Table
        The table of parameters for the ModelGrid
    inv_file: str
        An inventory file to more quickly load the database

    """
    def __init__(self, model_directory, bibcode='2013A & A...553A...6H',
                 names={'Teff': 'PHXTEFF', 'logg': 'PHXLOGG',
                        'FeH': 'PHXM_H', 'mass': 'PHXMASS', 'Lbol': 'PHXLUM'},
                 resolution=None, wave_units=q.um, **kwargs):
        """
        Initializes the model grid by creating a table with a column
        for each parameter and ingests the spectra

        Parameters
        ----------
        model_directory: str
            The path to the directory of FITS files of spectra,
            which may include a filename with a wildcard caharacter
        bibcode: str, array-like (optional)
            The bibcode or list of bibcodes for this data set
        names: dict (optional)
            A dictionary to rename the table columns. The Phoenix
            model keywords are given as an example
        resolution: int (optional)
            The desired wavelength resolution (lambda/d_lambda)
            of the grid spectra
        wave_units: astropy.units.quantity
        """
        # Make sure we can use glob if a directory
        # is given without a wildcard
        if '*' not in model_directory:
            model_directory = os.path.join(model_directory, '*')

        # Check for a precomputed pickle of this ModelGrid
        model_grid = None
        if model_directory.endswith('/*'):
            # Location of model_grid pickle
            file = model_directory.replace('*', 'model_grid.p')

            if os.path.isfile(file):
                model_grid = pickle.load(open(file, 'rb'))

                # Make sure the model_grid.path matches the given model_directory
                if os.path.dirname(file) != os.path.realpath(model_grid['path']):
                    _ = os.system('rm {}'.format(file))
                    flx_file = file.replace('model_grid.p', 'model_grid_flux.hdf5')

                    # Delete flux file if it exists
                    if os.path.isfile(flx_file):
                        _ = os.system('rm {}'.format(flx_file))

                    # Set model_grid to None so it regenerates it with the correct path
                    model_grid = None

        # Instantiate the precomputed model grid
        if model_grid is not None:

            for k, v in model_grid.items():
                setattr(self, k, v)

            self.flux_file = os.path.join(self.path, 'model_grid_flux.hdf5')
            self.flux = None
            self.wavelength = None
            self.mu = None

            del model_grid

        # Or compute it from scratch
        else:

            # Print update...
            if model_directory.endswith('/*'):

                print("Indexing models...")

            # Create some attributes
            self.path = os.path.dirname(model_directory) + '/'
            self.refs = None
            self.wave_rng = (0 * q.um, 40 * q.um)
            self.flux_file = os.path.join(self.path, 'model_grid_flux.hdf5')
            self.flux = None
            self.wavelength = None
            self.mu = None

            # Save the refs to a References() object
            if bibcode:
                if isinstance(bibcode, (list, tuple)):
                    pass
                elif bibcode and isinstance(bibcode, str):
                    bibcode = [bibcode]
                else:
                    pass

                self.refs = bibcode
                # _check_for_ref_object()

            # Get list of spectral intensity files
            files = glob(model_directory)
            filenames = []
            if not files:
                print('No files match', model_directory, '.')
                return

            # Parse the FITS headers
            vals, dtypes = [], []
            for f in files:
                if f.endswith('.fits'):
                    try:
                        header = fits.getheader(f)
                        keys = np.array(header.cards).T[0]
                        dtypes = [type(i[1]) for i in header.cards]
                        vals.append([header.get(k) for k in keys])
                        filenames.append(f.split('/')[-1])
                    except Exception:
                        print(f, 'could not be read into the model grid.')

            # Fix data types, trim extraneous values, and make the table
            dtypes = [str if d == bool else d for d in dtypes]
            vals = [v[: len(dtypes)] for v in vals]
            table = at.Table(np.array(vals), names=keys, dtype=dtypes)

            # Add the filenames as a column
            table['filename'] = filenames

            # Rename any columns
            for new, old in names.items():
                try:
                    table.rename_column(old, new)
                except Exception:
                    print('No column named', old)

            # Remove columns where the values are all the same
            # and store value as attribute instead
            for n in table.colnames:
                val = table[n][0]
                exc = n not in ['Teff', 'logg', 'FeH']
                if list(table[n]).count(val) == len(table[n]) and exc:
                    setattr(self, n, val)
                    table.remove_column(n)

            # Store the table in the data attribute
            self.data = table

            # Store the parameter ranges
            self.Teff_vals = np.asarray(np.unique(table['Teff']))
            self.logg_vals = np.asarray(np.unique(table['logg']))
            self.FeH_vals = np.asarray(np.unique(table['FeH']))

            # Write an inventory file to this directory for future table loads
            if model_directory.endswith('/*'):
                self.file = file
                try:
                    pickle.dump(self.__dict__, open(self.file, 'wb'))
                except IOError:
                    print('Could not write model grid to', self.file)

        # Print something
        print(len(self.data), 'models loaded from', self.path)

        # In case no filter is used
        self.n_bins = 1

        # Set the wavelength_units
        self.wave_units = q.AA
        if wave_units:
            self.set_units(wave_units)
        else:
            self.const = 1

        # Save the desired resolution
        self.resolution = resolution

        # Customize from the get-go
        if kwargs:
            self.customize(**kwargs)

    def export(self, filepath, **kwargs):
        """Export the model with the given parameters to a FITS file
        at the given filepath

        Parameters
        ----------
        filepath: str
            The path to the target FITS file
        """
        if not filepath.endswith('.fits'):
            raise IOError("Target file must have a .fits extension.")

        # Get the model
        model = self.get(**kwargs)

        # Get a dummy FITS file
        ffile = resource_filename('ExoCTK', 'data/core/ModelGrid_tmp.fits')
        hdu = fits.open(ffile)

        # Replace the data
        hdu[0].data = model['flux']
        hdu[1].data = model['mu']
        hdu[0].header['PHXTEFF'] = model['Teff']
        hdu[0].header['PHXLOGG'] = model['logg']
        hdu[0].header['PHXM_H'] = model['FeH']

        # Update the wavelength
        wave = model['wave']
        hdu[0].header['CRVAL1'] = min(wave)
        hdu[0].header['CDELT1'] = np.mean(np.diff(wave))
        hdu[0].header['CUNIT1'] = 'Micron'

        # Write the file
        hdu.writeto(filepath)

    def get(self, Teff, logg, FeH, resolution=None, interp=True):
        """
        Retrieve the wavelength, flux, and effective radius
        for the spectrum of the given parameters

        Parameters
        ----------
        Teff: int
            The effective temperature (K)
        logg: float
            The logarithm of the surface gravity (dex)
        FeH: float
            The logarithm of the ratio of the metallicity
            and solar metallicity (dex)
        resolution: int (optional)
            The desired wavelength resolution (lambda/d_lambda)
        interp: bool
            Interpolate the model if possible

        Returns
        -------
        dict
            A dictionary of arrays of the wavelength, flux, and
            mu values and the effective radius for the given model

        """
        # See if the model with the desired parameters is witin the grid
        in_grid = all([(Teff >= min(self.Teff_vals)) &
                       (Teff <= max(self.Teff_vals)) &
                       (logg >= min(self.logg_vals)) &
                       (logg <= max(self.logg_vals)) &
                       (FeH >= min(self.FeH_vals)) &
                       (FeH <= max(self.FeH_vals))])

        if in_grid:

            # See if the model with the desired parameters is a true grid point
            on_grid = self.data[[(self.data['Teff'] == Teff) &
                                 (self.data['logg'] == logg) &
                                 (self.data['FeH'] == FeH)]] in self.data

            # Grab the data if the point is on the grid
            if on_grid:

                # Get the row index and filepath
                row, = np.where((self.data['Teff'] == Teff) &
                                (self.data['logg'] == logg) &
                                (self.data['FeH'] == FeH))[0]

                filepath = self.path + str(self.data[row]['filename'])

                # Get the flux, mu, and abundance arrays
                raw_flux = fits.getdata(filepath, 0)
                mu = fits.getdata(filepath, 1)
                # abund = fits.getdata(filepath, 2)

                # Construct full wavelength scale and convert to microns
                if self.CRVAL1 == '-':
                    # Try to get data from WAVELENGTH extension...
                    dat = fits.getdata(filepath, ext=-1)
                    raw_wave = np.array(dat).squeeze()
                else:
                    # ...or try to generate it
                    b = self.CDELT1 * np.arange(len(raw_flux[0]))
                    raw_wave = np.array(self.CRVAL1 + b).squeeze()

                # Convert from A to desired units
                raw_wave *= self.const

                # Janky unit nullification
                def toQ(val):
                    return val if hasattr(val, 'unit') else val * self.wave_units

                # Trim the wavelength and flux arrays
                idx, = np.where(np.logical_and(raw_wave * self.wave_units >= toQ(self.wave_rng[0]),
                                               raw_wave * self.wave_units <= toQ(self.wave_rng[1])))
                flux = raw_flux[:, idx]
                wave = raw_wave[idx]

                # Bin the spectrum if necessary
                if resolution is not None or self.resolution is not None:

                    # Calculate zoom
                    z = utils.calc_zoom(resolution or self.resolution, wave)
                    wave = zoom(wave, z)
                    flux = zoom(flux, (1, z))

                # Make a dictionary of parameters
                # This should really be a core.Spectrum() object!
                row_data = self.data[row].as_void()
                spec_dict = dict(zip(self.data.colnames, row_data))
                spec_dict['wave'] = wave
                spec_dict['flux'] = flux
                spec_dict['mu'] = mu

            # If not on the grid, interpolate to it
            else:
                # Call grid_interp method
                if interp:
                    spec_dict = self.grid_interp(Teff, logg, FeH)
                else:
                    return

            return spec_dict

        else:
            print('Teff: ', Teff, ' logg: ', logg, ' FeH: ', FeH,
                  ' model not in grid.')
            return

    def grid_interp(self, Teff, logg, FeH, plot=False):
        """
        Interpolate the grid to the desired parameters

        Parameters
        ----------
        Teff: int
            The effective temperature (K)
        logg: float
            The logarithm of the surface gravity (dex)
        FeH: float
            The logarithm of the ratio of the metallicity
            and solar metallicity (dex)
        plot: bool
            Plot the interpolated spectrum along
            with the 8 neighboring grid spectra

        Returns
        -------
        dict
            A dictionary of arrays of the wavelength, flux, and
            mu values and the effective radius for the given model
        """
        # Load the fluxes
        if self.flux is None:
            self.load_flux()

        # Get the flux array
        flux = self.flux.copy()

        # Get the interpolable parameters
        params, values = [], []
        for p, v in zip([self.Teff_vals, self.logg_vals, self.FeH_vals],
                        [Teff, logg, FeH]):
            if len(p) > 1:
                params.append(p)
                values.append(v)
        values = np.asarray(values)
        label = '{}/{}/{}'.format(Teff, logg, FeH)

        try:
            # Interpolate flux values at each wavelength
            # using a pool for multiple processes
            print('Interpolating grid point [{}]...'.format(label))
            processes = 8
            mu_index = range(flux.shape[-2])
            start = time.time()
            pool = multiprocessing.Pool(processes)
            func = partial(utils.interp_flux, flux=flux, params=params,
                           values=values)
            new_flux, generators = zip(*pool.map(func, mu_index))
            pool.close()
            pool.join()

            # Clean up and time of execution
            new_flux = np.asarray(new_flux)
            generators = np.asarray(generators)
            print('Run time in seconds: ', time.time() - start)

            # Interpolate mu value
            interp_mu = RegularGridInterpolator(params, self.mu)
            mu = interp_mu(np.array(values)).squeeze()

            # Make a dictionary to return
            grid_point = {'Teff': Teff, 'logg': logg, 'FeH': FeH,
                          'mu': mu, 'flux': new_flux, 'wave': self.wavelength,
                          'generators': generators}

            return grid_point

        except IOError:
            print('Grid too sparse. Could not interpolate.')
            return

    def load_flux(self, reset=False):
        """
        Retrieve the flux arrays for all models
        and load into the ModelGrid.array attribute
        with shape (Teff, logg, FeH, mu, wavelength)
        """
        if reset:

            # Delete the old file and clear the flux attribute
            if os.path.isfile(self.flux_file):
                os.remove(self.flux_file)
            self.flux = None

        if self.flux is None:

            print('Loading flux into table...')

            if os.path.isfile(self.flux_file):

                # Load the flux from the HDF5 file
                f = h5py.File(self.flux_file, "r")
                self.flux = f['flux'][:]
                self.mu = f['mu'][:]
                self.wavelength = f['wave'][:]
                f.close()

            else:

                # Get array dimensions
                T, G, M = self.Teff_vals, self.logg_vals, self.FeH_vals
                shp = [len(T), len(G), len(M)]
                n, N = 1, np.prod(shp)

                # Iterate through rows
                for nt, teff in enumerate(T):
                    for ng, logg in enumerate(G):
                        for nm, feh in enumerate(M):

                            try:

                                # Retrieve flux using the `get()` method
                                d = self.get(teff, logg, feh, interp=False)

                                if d:

                                    # Make sure arrays exist
                                    if self.flux is None:
                                        new_shp = shp + list(d['flux'].shape)
                                        self.flux = np.zeros(new_shp)
                                    if self.mu is None:
                                        new_shp = shp + list(d['mu'].shape)
                                        self.mu = np.zeros(new_shp)

                                    # Add data to respective arrays
                                    self.flux[nt, ng, nm] = d['flux']
                                    self.mu[nt, ng, nm] = d['mu'].squeeze()

                                    # Get the wavelength array
                                    if self.wavelength is None:
                                        self.wavelength = d['wave']

                                    # Garbage collection
                                    del d

                                    # Print update
                                    n += 1
                                    msg = "{: .2f}% complete.".format(n * 100. / N)
                                    print(msg, end='\r')

                            except IOError:
                                # No model computed so reduce total
                                N -= 1

                # Load the flux into an HDF5 file
                f = h5py.File(self.flux_file, "w")
                f.create_dataset('flux', data=self.flux)
                f.create_dataset('mu', data=self.mu)
                f.create_dataset('wave', data=self.wavelength)
                f.close()
                # del dset
                print("100.00 percent complete!", end='\n')

        else:
            print('Data already loaded.')

    def customize(self, Teff_rng=(2300, 8000), logg_rng=(0, 6),
                  FeH_rng=(-2, 1), wave_rng=(0 * q.um, 40 * q.um), n_bins=''):
        """
        Trims the model grid by the given ranges in effective temperature,
        surface gravity, and metallicity. Also sets the wavelength range
        and number of bins for retrieved model spectra.

        Parameters
        ----------
        Teff_rng: array-like
            The lower and upper inclusive bounds for the effective
            temperature (K)
        logg_rng: array-like
            The lower and upper inclusive bounds for the logarithm of the
            surface gravity (dex)
        FeH_rng: array-like
            The lower and upper inclusive bounds for the logarithm of the
            ratio of the metallicity and solar metallicity (dex)
        wave_rng: array-like
            The lower and upper inclusive bounds for the wavelength (microns)
        n_bins: int
            The number of bins for the wavelength axis

        """
        # Make a copy of the grid
        grid = self.data.copy()
        self.wave_rng = wave_rng
        self.n_bins = n_bins or self.n_bins

        # Filter grid by given parameters
        self.data = grid[[(grid['Teff'] >= Teff_rng[0]) &
                          (grid['Teff'] <= Teff_rng[1]) &
                          (grid['logg'] >= logg_rng[0]) &
                          (grid['logg'] <= logg_rng[1]) &
                          (grid['FeH'] >= FeH_rng[0]) &
                          (grid['FeH'] <= FeH_rng[1])]]

        # Print a summary of the returned grid
        print('{}/{}'.format(len(self.data), len(grid)),
              'spectra in parameter range',
              'Teff: ', Teff_rng, ', logg: ', logg_rng,
              ', FeH: ', FeH_rng, ', wavelength: ', wave_rng)

        # Do nothing if he cut leaves the grid empty
        if len(self.data) == 0:
            self.data = grid
            print('The given param ranges would leave 0 models in the grid.')
            print('The model grid has not been updated. Please try again.')
            return

        # Update the wavelength and flux attributes
        if isinstance(self.wavelength, np.ndarray):
            w = self.wavelength
            W_idx, = np.where((w >= wave_rng[0]) & (w <= wave_rng[1]))
            T_idx, = np.where((self.Teff_vals >= Teff_rng[0]) &
                              (self.Teff_vals <= Teff_rng[1]))
            G_idx, = np.where((self.logg_vals >= logg_rng[0]) &
                              (self.logg_vals <= logg_rng[1]))
            M_idx, = np.where((self.FeH_vals >= FeH_rng[0]) &
                              (self.FeH_vals <= FeH_rng[1]))

            # Trim arrays
            self.wavelength = w[W_idx]
            self.flux = self.flux[T_idx[0]: T_idx[-1] + 1,
                                  G_idx[0]: G_idx[-1] + 1,
                                  M_idx[0]: M_idx[-1] + 1,
                                  :, W_idx[0]: W_idx[-1] + 1]
            self.mu = self.mu[T_idx[0]: T_idx[-1] + 1,
                              G_idx[0]: G_idx[-1] + 1,
                              M_idx[0]: M_idx[-1] + 1]

        # Update the parameter attributes
        self.Teff_vals = np.unique(self.data['Teff'])
        self.logg_vals = np.unique(self.data['logg'])
        self.FeH_vals = np.unique(self.data['FeH'])

        # Reload the flux array with the new grid parameters
        self.load_flux(reset=True)

        # Clear the grid copy from memory
        del grid

    def info(self):
        """
        Print a table of info about the current ModelGrid
        """
        # Get the info from the class
        tp = (int, bytes, bool, str, float, tuple, list, np.ndarray)
        info = [[k, str(v)] for k, v in vars(self).items()
                if isinstance(v, tp)]

        # Make the table
        table = at.Table(np.asarray(info).reshape(len(info), 2),
                         names=['Attributes', 'Values'])

        # Sort and print
        table.sort('Attributes')
        table.pprint(max_width=-1, align=['>', '<'])

    def reset(self):
        """
        Reset the current grid to the original state
        """
        file = os.path.join(self.path + 'model_grid_flux.hdf5')

        if os.path.isfile(file):
            os.remove(file)

        self.__init__(self.path)

    def set_units(self, wave_units=q.um):
        """
        Set the wavelength and flux units

        Parameters
        ----------
        wave_units: str, astropy.units.core.PrefixUnit/CompositeUnit
            The wavelength units
        """
        # Set wavelength units
        old_unit = self.wave_units
        self.wave_units = q.Unit(wave_units)

        # Update the wavelength
        self.const = (old_unit / self.wave_units).decompose()._scale


class ACES(ModelGrid):
    """A convenience function to load the ACES model grid from the EXOCTK_DATA directory"""
    def __init__(self, **kwargs):
        """Initialize the ModelGrid object with the ACES models"""
        # Get the ACES model directory from the EXOCTK_DATA directory
        moddir = os.path.join(os.environ.get('EXOCTK_DATA'), 'modelgrid/ACES')

        # Initialize base class
        super().__init__(model_directory=moddir, **kwargs)


class ATLAS9(ModelGrid):
    """A convenience function to load the ATLAS9 model grid from the EXOCTK_DATA directory"""
    def __init__(self, **kwargs):
        """Initialize the ModelGrid object with the ACES models"""
        # Get the ACES model directory from the EXOCTK_DATA directory
        moddir = os.path.join(os.environ.get('EXOCTK_DATA'), 'modelgrid/ATLAS9')

        # Initialize base class
        super().__init__(model_directory=moddir, **kwargs)



#!/usr/bin/python
# -*- coding: latin-1 -*-
"""
A module to calculate limb darkening coefficients from a grid of model spectra
"""
import copy
import inspect
import os
import warnings

from astropy.io import ascii as ii
import astropy.table as at
import astropy.units as q
from astropy.utils.exceptions import AstropyWarning
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import rc
import numpy as np
from scipy.optimize import curve_fit
from svo_filters import svo
import bokeh.plotting as bkp
from bokeh.models import Range1d
from bokeh.models.widgets import Panel, Tabs

from .. import utils
from .. import modelgrid

rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica'], 'size': 16})
rc('text', usetex=True)

warnings.simplefilter('ignore', category=AstropyWarning)
warnings.simplefilter('ignore', category=FutureWarning)


def ld_profile(name='quadratic', latex=False):
    """
    Define the function to fit the limb darkening profile

    Reference:
        https://www.cfa.harvard.edu/~lkreidberg/batman/
        tutorial.html#limb-darkening-options

    Parameters
    ----------
    name: str
        The name of the limb darkening profile function to use,
        including 'uniform', 'linear', 'quadratic', 'square-root',
        'logarithmic', 'exponential', '3-parameter', and '4-parameter'
    latex: bool
        Return the function as a LaTeX formatted string

    Returns
    -------
    function, str
        The corresponding function for the given profile

    """
    # Supported profiles a la BATMAN
    names = ['uniform', 'linear', 'quadratic', 'square-root', 'logarithmic', 'exponential', '3-parameter', '4-parameter']

    # Check that the profile is supported
    if name in names:

        # Uniform
        if name == 'uniform':
            def profile(m, c1):
                return c1

        # Linear
        if name == 'linear':
            def profile(m, c1):
                return 1. - c1 * (1. - m)

        # Quadratic
        if name == 'quadratic':
            def profile(m, c1, c2):
                return 1. - c1 * (1. - m) - c2 * (1. - m)**2

        # Square-root
        if name == 'square-root':
            def profile(m, c1, c2):
                return 1. - c1 * (1. - m) - c2 * (1. - np.sqrt(m))

        # Logarithmic
        if name == 'logarithmic':
            def profile(m, c1, c2):
                return 1. - c1 * (1. - m) - c2 * m * np.log(m)

        # Exponential
        if name == 'exponential':
            def profile(m, c1, c2):
                return 1. - c1 * (1. - m) - c2 / (1. - np.e**m)

        # 3-parameter
        if name == '3-parameter':
            def profile(m, c1, c2, c3):
                return 1. - c1 * (1. - m) - c2 * (1. - m**1.5) - c3 * (1. - m**2)

        # 4-parameter
        if name == '4-parameter':
            def profile(m, c1, c2, c3, c4):
                return 1. - c1 * (1. - m**0.5) - c2 * (1. - m) - c3 * (1. - m**1.5) - c4 * (1. - m**2)

        if latex:
            profile = inspect.getsource(profile).replace('\n', '')
            profile = profile.replace('\\', '').split('return ')[1]

            for i, j in [('**', '^'), ('m', '\mu'), (' ', ''), ('np.', '\\'), ('0.5', '{0.5}'), ('1.5', '{1.5}')]:
                profile = profile.replace(i, j)

        return profile

    else:
        print("'{}' is not a supported profile. Try".format(name), names)
        return


class LDC:
    """A class to hold all the LDCs you want to run

    Example
    -------
    from exoctk.limb_darkening import limb_darkening_fit as lf
    from svo_filters import Filter
    ld = lf.LDC(model_grid='ACES')
    bp = Filter('WFC3_IR.G141', n_bins=5)
    ld.calculate(4000, 4.5, 0.0, 'quadratic', bandpass=bp)
    ld.calculate(4000, 4.5, 0.0, '4-parameter', bandpass=bp)
    ld.plot(show=True)
    """
    def __init__(self, model_grid='ACES'):
        """Initialize an LDC object

        Parameters
        ----------
        model_grid: exoctk.modelgrid.ModelGrid
            The grid of synthetic spectra from which the coefficients will
            be calculated
        """
        # Try ACES or ATLAS
        if model_grid == 'ACES':
            model_grid = modelgrid.ACES(resolution=700)
        if model_grid == 'ATLAS9':
            model_grid = modelgrid.ATLAS9(resolution=700)

        # Make sure it's a ModelGrid object
        if not isinstance(model_grid, modelgrid.ModelGrid):
            raise TypeError("'model_grid' must be a exoctk.modelgrid.ModelGrid object.")

        # Load the model grid
        self.model_grid = model_grid

        # Table for results
        columns = ['name', 'Teff', 'logg', 'FeH', 'profile', 'filter', 'models', 'coeffs', 'errors', 'wave', 'wave_min', 'wave_eff', 'wave_max', 'scaled_mu', 'raw_mu', 'mu_min', 'scaled_ld', 'raw_ld', 'ld_min', 'ldfunc', 'flux', 'bandpass', 'color']
        dtypes = ['|S20', float, float, float, '|S20', '|S20', '|S20', object, object, object, np.float16, np.float16, np.float16, object, object, np.float16, object, object, np.float16, object, object, object, '|S20']
        self.results = at.Table(names=columns, dtype=dtypes)

        self.ld_color = {'quadratic': 'blue', '4-parameter': 'red', 'exponential': 'green', 'linear': 'orange', 'square-root': 'cyan', '3-parameter': 'magenta', 'logarithmic': 'pink', 'uniform': 'purple'}

        self.count = 1

    @staticmethod
    def bootstrap_errors(mu_vals, func, coeffs, errors, n_samples=1000):
        """
        Bootstrapping LDC errors

        Parameters
        ----------
        mu_vals: sequence
            The mu values
        func: callable
            The LD profile function
        coeffs: sequence
            The coefficients
        errors: sequence
            The errors on each coeff
        n_samples: int
            The number of samples

        Returns
        -------
        tuple
            The lower and upper errors
        """
        # Generate n_samples
        vals = []
        for n in range(n_samples):
            co = np.random.normal(coeffs, errors)
            vals.append(func(mu_vals, *co))

        # r = np.array(list(zip(*vals)))
        dn_err = np.min(np.asarray(vals), axis=0)
        up_err = np.max(np.asarray(vals), axis=0)

        return dn_err, up_err

    def calculate(self, Teff, logg, FeH, profile, mu_min=0.05, ld_min=0.01,
                  bandpass=None, name=None, color=None, **kwargs):
        """
        Calculates the limb darkening coefficients for a given synthetic
        spectrum. If the model grid does not contain a spectrum of the given
        parameters, the grid is interpolated to those parameters.

        Reference for limb-darkening laws:
        http://www.astro.ex.ac.uk/people/sing/David_Sing/Limb_Darkening.html

        Parameters
        ----------
        Teff: int
            The effective temperature of the model
        logg: float
            The logarithm of the surface gravity
        FeH: float
            The logarithm of the metallicity
        profile: str
            The name of the limb darkening profile function to use,
            including 'uniform', 'linear', 'quadratic', 'square-root',
            'logarithmic', 'exponential', and '4-parameter'
        mu_min: float
            The minimum mu value to consider
        ld_min: float
            The minimum limb darkening value to consider
        bandpass: svo_filters.svo.Filter() (optional)
            The photometric filter through which the limb darkening
            is to be calculated
        name: str (optional)
            A name for the calculation
        color: str (optional)
            A color for the plotted result
        """
        # Define the limb darkening profile function
        ldfunc = ld_profile(profile)

        if not ldfunc:
            raise ValueError("No such LD profile:", profile)

        # Get the grid point
        grid_point = self.model_grid.get(Teff, logg, FeH)

        # Retrieve the wavelength, flux, mu, and effective radius
        wave = grid_point.get('wave')
        flux = grid_point.get('flux')
        mu = grid_point.get('mu').squeeze()

        # Use tophat oif no bandpass
        if bandpass is None:
            units = self.model_grid.wave_units
            bandpass = svo.Filter('tophat', wave_min=np.min(wave) * units, wave_max=np.max(wave) * units)

        # Check if a bandpass is provided
        if not isinstance(bandpass, svo.Filter):
            raise TypeError("Invalid bandpass of type", type(bandpass))

        # # Make sure the bandpass has coverage
        # bp_min = bandpass.wave_min.value
        # bp_max = bandpass.wave_max.value
        # mg_min = self.model_grid.wave_rng[0].value
        # mg_max = self.model_grid.wave_rng[-1].value
        # if bp_min < mg_min or bp_max > mg_max:
        #     raise ValueError('Bandpass {} not covered by model grid of\
        #                       wavelength range {}'.format(bandpass.filterID,
        #                                                   self.model_grid
        #                                                       .wave_rng))

        # Apply the filter
        try:
            flux, _ = bandpass.apply([wave, flux])  # Sometimes this returns a tuple
        except ValueError:
            flux = bandpass.apply([wave, flux])  # Sometimes it returns one value

        # Make rsr curve 3 dimensions if there is only one
        # wavelength bin, then get wavelength only
        bp = bandpass.rsr
        if bp.ndim == 2:
            bp = bp[None, :]
        wave = bp[:, 0, :]

        # Calculate mean intensity vs. mu
        wave = wave[None, :] if wave.ndim == 1 else wave
        flux = flux[None, :] if flux.ndim == 2 else flux
        mean_i = np.nanmean(flux, axis=-1)
        mean_i[mean_i == 0] = np.nan

        # Calculate limb darkening, I[mu]/I[1] vs. mu
        ld = mean_i / mean_i[:, np.where(mu == max(mu))].squeeze(axis=-1)

        # Rescale mu values to make f(mu=0)=ld_min
        # for the case where spherical models extend beyond limb
        ld_avg = np.nanmean(ld, axis=0)
        muz = np.interp(ld_min, ld_avg, mu) if any(ld_avg < ld_min) else 0
        mu = (mu - muz) / (1 - muz)

        # Trim to useful mu range
        imu, = np.where(mu > mu_min)
        scaled_mu, scaled_ld = mu[imu], ld[:, imu]

        # Fit limb darkening coefficients for each wavelength bin
        for n, ldarr in enumerate(scaled_ld):

            # Get effective wavelength of bin
            wave_eff = bandpass.centers[0, n].round(5)

            try:

                # Fit polynomial to data
                coeffs, cov = curve_fit(ldfunc, scaled_mu, ldarr, method='lm')

                # Calculate errors from covariance matrix diagonal
                errs = np.sqrt(np.diag(cov))

                # Make a dictionary or the results
                result = {}

                # Check the count
                result['name'] = name or 'Calculation {}'.format(self.count)
                self.count += 1
                if len(bandpass.centers[0]) == len(scaled_ld) and name is None:
                    result['name'] = '{} {}'.format(str(round(bandpass.centers[0][n], 2)), self.model_grid.wave_units)

                # Set a color if possible
                result['color'] = color or self.ld_color[profile]

                # Add the results
                result['Teff'] = Teff
                result['logg'] = logg
                result['FeH'] = FeH
                result['filter'] = bandpass.filterID
                result['models'] = self.model_grid.path
                result['raw_mu'] = mu
                result['raw_ld'] = ld[n]
                result['scaled_mu'] = scaled_mu
                result['scaled_ld'] = ldarr
                result['flux'] = flux[n]
                result['wave'] = wave[n]
                result['mu_min'] = mu_min
                result['bandpass'] = bandpass
                result['ldfunc'] = ldfunc
                result['coeffs'] = coeffs
                result['errors'] = errs
                result['profile'] = profile
                result['n_bins'] = bandpass.n_bins
                result['pixels_per_bin'] = bandpass.pixels_per_bin
                result['wave_min'] = wave[n, 0].round(5)
                result['wave_eff'] = wave_eff
                result['wave_max'] = wave[n, -1].round(5)

                # Add the coeffs
                for n, (coeff, err) in enumerate(zip(coeffs, errs)):
                    cname = 'c{}'.format(n + 1)
                    ename = 'e{}'.format(n + 1)
                    result[cname] = coeff.round(3)
                    result[ename] = err.round(3)

                    # Add the coefficient column to the table if not present
                    if cname not in self.results.colnames:
                        self.results[cname] = [np.nan] * len(self.results)
                        self.results[ename] = [np.nan] * len(self.results)

                # Add the new row to the table
                result = {i: j for i, j in result.items() if i in self.results.colnames}
                self.results.add_row(result)

            except ValueError:
                print("Could not calculate coefficients at {}".format(wave_eff))

    def plot_tabs(self, show=False, **kwargs):
        """Plot the LDCs in a tabbed figure

        Parameters
        ----------
        fig: matplotlib.pyplot.figure, bokeh.plotting.figure (optional)
            An existing figure to plot on
        show: bool
            Show the figure
        """
        # Change names to reflect ld profile
        old_names = self.results['name']
        for n, row in enumerate(self.results):
            self.results[n]['name'] = row['profile']

        # Draw a figure for each wavelength bin
        tabs = []
        for wav in np.unique(self.results['wave_eff']):

            # Plot it
            TOOLS = 'box_zoom, box_select, crosshair, reset, hover'
            fig = bkp.figure(tools=TOOLS, x_range=Range1d(0, 1), y_range=Range1d(0, 1), plot_width=800, plot_height=400)
            self.plot(wave_eff=wav, fig=fig)

            # Plot formatting
            fig.legend.location = 'bottom_right'
            fig.xaxis.axis_label = 'mu'
            fig.yaxis.axis_label = 'Intensity'

            tabs.append(Panel(child=fig, title=str(wav)))

        # Make the final tabbed figure
        final = Tabs(tabs=tabs)

        # Put the names back
        self.results['name'] = old_names

        if show:
            bkp.show(final)
        else:
            return final

    def plot(self, fig=None, show=False, **kwargs):
        """Plot the LDCs

        Parameters
        ----------
        fig: matplotlib.pyplot.figure, bokeh.plotting.figure (optional)
            An existing figure to plot on
        show: bool
            Show the figure
        """
        # Separate plotting kwargs from parameter kwargs
        pwargs = {i: j for i, j in kwargs.items() if i in self.results.columns}
        kwargs = {i: j for i, j in kwargs.items() if i not in pwargs.keys()}

        # Filter the table by given kwargs
        table = utils.filter_table(self.results, **pwargs)

        for row in table:

            # Set color and label for plot
            color = row['color']
            label = row['name']

            # Generate smooth curve
            ldfunc = row['ldfunc']
            mu_vals = np.linspace(0, 1, 1000)
            ld_vals = ldfunc(mu_vals, *row['coeffs'])

            # Generate smooth errors
            dn_err, up_err = self.bootstrap_errors(mu_vals, ldfunc, row['coeffs'], row['errors'])

            # Matplotlib fig by default
            if fig is None:
                fig = bkp.figure()

            # Add fits to matplotlib
            if isinstance(fig, matplotlib.figure.Figure):

                # Make axes
                ax = fig.add_subplot(111)

                # Plot the fitted points
                ax.errorbar(row['raw_mu'], row['raw_ld'], c='k', ls='None', marker='o', markeredgecolor='k', markerfacecolor='None')

                # Plot the mu cutoff
                ax.axvline(row['mu_min'], color='0.5', ls=':')

                # Draw the curve and error
                ax.plot(mu_vals, ld_vals, color=color, label=label, **kwargs)
                ax.fill_between(mu_vals, dn_err, up_err, color=color, alpha=0.1)
                ax.set_ylim(0, 1)
                ax.set_xlim(0, 1)

            # Or to bokeh!
            else:

                # Set the plot elements
                fig.x_range = Range1d(0, 1)
                fig.y_range = Range1d(0, 1)
                fig.xaxis.axis_label = 'mu'
                fig.yaxis.axis_label = 'Normalized Intensity'
                fig.legend.location = "bottom_right"

                # Plot the fitted points
                fig.circle(row['raw_mu'], row['raw_ld'], fill_color='black')

                # Plot the mu cutoff
                fig.line([row['mu_min']] * 2, [0, 1], legend_label='cutoff', line_color='#6b6ecf', line_dash='dotted')

                # Draw the curve and error
                fig.line(mu_vals, ld_vals, line_color=color, legend_label=label, **kwargs)
                vals = np.append(mu_vals, mu_vals[::-1])
                evals = np.append(dn_err, up_err[::-1])
                fig.patch(vals, evals, color=color, fill_alpha=0.2, line_alpha=0)

        if show:
            if isinstance(fig, matplotlib.figure.Figure):
                plt.xlabel('$\mu$')
                plt.ylabel('$I(\mu)/I(\mu = 1)$')
                plt.legend(loc=0, frameon=False)
                plt.show()
            else:
                bkp.show(fig)

        else:
            return fig

    def save(self, filepath):
        """
        Save the LDC results to file

        Parameters
        ----------
        filepath: str
            The complete filepath to save the results to
        """
        # Make sure it is a string
        if not isinstance(filepath, str):
            raise TypeError("{}: Expecting a string for 'filepath' argument".format(type(filepath)))

        # Make sure it's a valid path
        if not os.path.exists(os.path.dirname(filepath)):
            raise ValueError("{}: Not a valid path")

        # Copy the results table
        keep_cols = ['Teff', 'logg', 'FeH', 'profile', 'filter', 'models', 'wave_min', 'wave_eff', 'wave_max', 'c1', 'e1', 'c2', 'e2', 'c3', 'e3', 'c4', 'e4']
        results = self.results[keep_cols]

        # Save to file
        ii.write(results, filepath, format='fixed_width_two_line')



#! /usr/bin/env python
# Version 1. August 2, 2010
# Version 2. August 3, 2010
#   Got rid of degrees trig functions
# Joe Filippazzo, 2018/06/26, Made PEP8 compliant

from math import cos, sin, atan2, acos, pi

from .math_extensionsx import sind, cosd

D2R = pi / 180.
R2D = 180. / pi
PI2 = 2. * pi
epsilon = 23.43929 * D2R  # obliquity of the ecliptic J2000


def pa(tgt_c1, tgt_c2, obj_c1, obj_c2):
    """Calculates position angle of object at tgt position.

    Parameters
    ----------
    tgt_c1: float
        The RA of the target.
    tgt_c2: float
        The Dec of the target.
    obj_c1: float
        The RA of the reference.
    obj_c2: float
        The Dec of the reference.

    Returns
    -------
    float
        The position angle.
    """
    y = cos(obj_c2) * sin(obj_c1 - tgt_c1)
    c = cos(obj_c2) * sin(tgt_c2) * cos(obj_c1 - tgt_c1)
    x = sin(obj_c2) * cos(tgt_c2) - c
    p = atan2(y, x)
    if p < 0.:
        p += PI2
    if p >= PI2:
        p -= PI2
    return p


def delta_pa_no_roll(pos1_c1, pos1_c2, pos2_c1, pos2_c2):
    """Calculates the change in position angle between two positions with no
    roll about V1

    Parameters
    ----------
    pos1_c1: float
        The RA of the first position.
    pos1_c2: float
        The Dec of the first position.
    pos2_c1: float
        The RA of the second position.
    pos2_c2: float
        The Dec of the second position.

    Returns
    -------
    float
        The change in position angle.
    """
    u = (sin(pos1_c2) + sin(pos2_c2)) * sin(pos2_c1 - pos1_c1)
    a = sin(pos2_c2) * cos(pos2_c1 - pos1_c1)
    v = cos(pos2_c1 - pos1_c1) + cos(pos1_c2) * cos(pos2_c2) + sin(pos1_c2) * a
    return atan2(u, v)


def dist(obj1_c1, obj1_c2, obj2_c1, obj2_c2):
    """Calculates the angular distance betrween two objects, positions
    specified in spherical coordinates

    Parameters
    ----------
    obj1_c1: float
        The RA of the first object.
    obj1_c2: float
        The Dec of the first object.
    obj2_c1: float
        The RA of the second object.
    obj2_c2: float
        The Dec of the second object.

    Returns
    -------
    float
        The distance between the objects.
    """
    b = sin(obj2_c2) * sin(obj1_c2)
    x = cos(obj2_c2) * cos(obj1_c2) * cos(obj2_c1 - obj1_c1) + b
    return acos(unit_limit(x))


def JWST_same_ori(tgt0_c1, tgt0_c2, p0, tgt_c1, tgt_c2):
    """Calculates normal orientation of second target, given first target's
    orientation is normal. This is in Ecliptic coordinates!

    Parameters
    ----------
    tgt0_c1: float
        The RA of the first target.
    tgt0_c2: float
        The Dec of the first target.
    p0: float
        The origin.
    tgt1_c1: float
        The RA of the second target.
    tgt1_c2: float
        The Dec of the second target.

    Returns
    -------
    float
        The normal orientation.
    """
    long_sun = atan2(-sind(p0) * sin(tgt0_c2), cosd(p0)) + tgt0_c1
    pp = atan2(-sin(long_sun - tgt_c1), cos(long_sun - tgt_c1) * sin(tgt_c2))
    if pp < 0.:
        pp += PI2
    return pp


def unit_limit(x):
    """ Forces value to be in [-1, 1]

    Parameters
    ----------
    x: float, int
        The value to adjust.

    Retruns
    -------
    float
        The adjusted value.
    """
    return min(max(-1., x), 1.)



import os
import pkg_resources
import sys

from astropy.io import fits
from bokeh.layouts import gridplot
from bokeh.plotting import figure
from bokeh.models import Range1d, LinearColorMapper, Label
from bokeh.models.widgets import Panel, Tabs
from bokeh.palettes import PuBu
import numpy as np

from . import visibilityPA as vpa

EXOCTK_DATA = os.environ.get('EXOCTK_DATA')
if not EXOCTK_DATA:
    print(
        'WARNING: The $EXOCTK_DATA environment variable is not set. Contamination overlap will not work. Please set the '
        'value of this variable to point to the location of the exoctk_data '
        'download folder.  Users may retreive this folder by clicking the '
        '"ExoCTK Data Download" button on the ExoCTK website, or by using '
        'the exoctk.utils.download_exoctk_data() function.')
    TRACES_PATH = None
else:
    TRACES_PATH = os.path.join(EXOCTK_DATA, 'exoctk_contam', 'traces')

disp_nircam = 0.001  # microns
lam0_nircam322w2 = 2.369
lam1_nircam322w2 = 4.417
lam0_nircam444w = 3.063
lam1_nircam444w = 5.111


def nirissContam(cube, paRange=[0, 360]):
    """ Generates the contamination figure that will be plotted on the website
    for NIRISS SOSS.
    """
    # Get data from FITS file
    if isinstance(cube, str):
        hdu = fits.open(cubeName)
        cube = hdu[0].data
        hdu.close()

    # Pull out the target trace and cube of neighbor traces
    trace1 = cube[0, :, :]
    trace2 = cube[1, :, :]
    cube = cube[2:, :, :]

    plotPAmin, plotPAmax = paRange

    # Start calculations
    if not TRACES_PATH:
        return None
    lam_file = os.path.join(TRACES_PATH, 'NIRISS', 'lambda_order1-2.txt')
    ypix, lamO1, lamO2 = np.loadtxt(lam_file, unpack=True)

    nPA = cube.shape[0]
    rows = cube.shape[1]
    cols = cube.shape[2]
    print('cols ', cols)
    dPA = 360 // nPA
    PA = np.arange(nPA) * dPA

    contamO1 = np.zeros([rows, nPA])
    contamO2 = np.zeros([rows, nPA])

    low_lim_col = 20
    high_lim_col = 41

    for row in np.arange(rows):
        # Contamination for order 1 of target trace
        i = np.argmax(trace1[row, :])
        tr = trace1[row, i - low_lim_col:i + high_lim_col]
        w = tr / np.sum(tr**2)
        ww = np.tile(w, nPA).reshape([nPA, tr.size])
        contamO1[row, :] = np.sum(
            cube[:, row, i - low_lim_col:i + high_lim_col] * ww, axis=1)

        # Contamination for order 2 of target trace
        if lamO2[row] < 0.6:
            continue
        i = np.argmax(trace2[row, :])
        tr = trace2[row, i - 20:i + 41]
        w = tr / np.sum(tr**2)
        ww = np.tile(w, nPA).reshape([nPA, tr.size])
        contamO2[row, :] = np.sum(cube[:, row, i - 20:i + 41] * ww, axis=1)

    return contamO1, contamO2


def nircamContam(cube, instrument, paRange=[0, 360]):
    """ Generates the contamination figure that will be plotted on the website
    for NIRCam Grism Time Series mode.

    PARAMETERS
    ----------
    cube : arr or str
        A 3D array of the simulated field at every Aperture Position Angle (APA).
        The shape of the cube is (361, subY, subX).
        or
        The name of an HDU .fits file sthat has the cube.

    instrument : str
        The name of the instrument + what filter is being used. For NIRCam the
        options are: 'NIRCam F322W2', 'NIRCam F444W'

    RETURNS
    -------
    bokeh plot
    """
    # Get data from FITS file
    if isinstance(cube, str):
        hdu = fits.open(cubeName)
        cube = hdu[0].data
        hdu.close()

    # Pull out the target trace and cube of neighbor traces
    targ = cube[0, :, :]  # target star order 1 trace
    # neighbor star order 1 and 2 traces in all the angles
    cube = cube[1:, :, :]

    # Remove background values < 1 as it can blow up contamination
    targ = np.where(targ < 1, 0, targ)

    PAmin, PAmax = paRange[0], paRange[1]
    PArange = np.arange(PAmin, PAmax, 1)

    nPA, rows, cols = cube.shape[0], cube.shape[1], cube.shape[2]

    contamO1 = np.zeros([nPA, cols])

    # the width of the trace (in Y-direction for NIRCam GTS)
    peak = targ.max()
    low_lim_row = np.where(targ > 0.0001 * peak)[0].min()
    high_lim_row = np.where(targ > 0.0001 * peak)[0].max()

    # the length of the trace (in X-direction for NIRCam GTS)
    targ_trace_start = np.where(targ > 0.0001 * peak)[1].min()
    targ_trace_stop = np.where(targ > 0.0001 * peak)[1].max()

    # Begin contam calculation at each channel (column) X
    for X in np.arange(cols):
        if (X < targ_trace_start) or (X > targ_trace_stop):
            continue

        peakY = np.argmax(targ[:, X])
        TOP, BOT = peakY + high_lim_row, peakY - low_lim_row

        tr = targ[BOT:TOP, X]

        # calculate weights
        wt = tr / np.sum(tr**2)
        ww = np.tile(wt, nPA).reshape([nPA, tr.size])

        contamO1[:, X] = np.sum(cube[:, BOT:TOP, X] * ww, axis=1)

    contamO1 = contamO1[:, targ_trace_start:targ_trace_stop]
    return contamO1


def miriContam(cube, paRange=[0, 360]):
    """ Generates the contamination figure that will be plotted on the website
    for MIRI LRS.
    """
    # Get data from FITS file
    if isinstance(cube, str):
        hdu = fits.open(cubeName)
        cube = hdu[0].data
        hdu.close()

    # Pull out the target trace and cube of neighbor traces
    targ = cube[0, :, :]  # target star order 1 trace
    # neighbor star order 1 and 2 traces in all the angles
    cube = cube[1:, :, :]

    # Remove background values < 1 as it can blow up contamination
    targ = np.where(targ < 1, 0, targ)

    PAmin, PAmax = paRange[0], paRange[1]
    PArange = np.arange(PAmin, PAmax, 1)

    nPA, rows, cols = cube.shape[0], cube.shape[1], cube.shape[2]

    contamO1 = np.zeros([rows, nPA])

    # the width of the trace (in Y-direction for NIRCam GTS)
    peak = targ.max()

    low_lim_col = np.where(targ > 0.0001 * peak)[1].min()
    high_lim_col = np.where(targ > 0.0001 * peak)[1].max()

    # the length of the trace (in X-direction for NIRCam GTS)
    targ_trace_start = np.where(targ > 0.0001 * peak)[0].min()
    targ_trace_stop = np.where(targ > 0.0001 * peak)[0].max()
    # Begin contam calculation at each channel (row) Y
    for Y in np.arange(rows):
        if (Y < targ_trace_start) or (Y > targ_trace_stop):
            continue

        peakX = np.argmax(targ[Y, :])
        LEFT, RIGHT = peakX - low_lim_col, peakX + high_lim_col

        tr = targ[Y, LEFT:RIGHT]

        # calculate weights
        wt = tr / np.sum(tr**2)
        ww = np.tile(wt, nPA).reshape([nPA, tr.size])

        contamO1[Y, :] = np.sum(cube[:, Y, LEFT:RIGHT] * wt,
                                where=~np.isnan(cube[:, Y, LEFT:RIGHT] * wt),
                                axis=1)

        #target = np.sum(cube[0, Y, LEFT:RIGHT], axis=0)

        # contamO1[Y, :] = np.sum(cube[:, Y, LEFT:RIGHT]*ww,
        #                        where=~np.isnan(cube[:, Y, LEFT:RIGHT]),
        #                        axis=1)#/target
    contamO1 = contamO1[targ_trace_start:targ_trace_stop, :]
    return contamO1


def contam(cube, instrument, targetName='noName', paRange=[0, 360],
           badPAs=np.asarray([]), tmpDir="", fig='', to_html=True):

    lam_file = os.path.join(TRACES_PATH, 'NIRISS', 'lambda_order1-2.txt')
    ypix, lamO1, lamO2 = np.loadtxt(lam_file, unpack=True)

    nPA, rows, cols = cube.shape[0], cube.shape[1], cube.shape[2]

    PAmin, PAmax = paRange[0], paRange[1]
    PA = np.arange(PAmin, PAmax, 1)

    # Generate the contam figure
    if instrument == 'NIRISS':
        contamO1, contamO2 = nirissContam(cube)
    elif (instrument == 'NIRCam F322W2') or (instrument == 'NIRCam F444W'):
        contamO1 = nircamContam(cube, instrument)
    elif instrument == 'MIRI':
        contamO1 = miriContam(cube)

    TOOLS = 'pan, box_zoom, crosshair, reset, hover'

    y = np.array([0., 0.])
    y1 = 0.07
    y2 = 0.12
    y3 = 0.17
    y4 = 0.23
    bad_PA_color = '#dddddd'
    bad_PA_alpha = 0.7
    dPA = 1

    # Order 1~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    # Contam plot
    if instrument == 'NIRISS':
        xlim0 = lamO1.min()
        xlim1 = lamO1.max()
    elif instrument == 'NIRCam F322W2':
        xlim0 = lam0_nircam322w2
        xlim1 = lam1_nircam322w2
    elif instrument == 'NIRCam F444W':
        xlim0 = lam0_nircam444w
        xlim1 = lam1_nircam444w
    elif instrument == 'MIRI':
        xlim0 = 5
        xlim1 = 12

    ylim0 = PAmin - 0.5
    ylim1 = PAmax + 0.5
    color_mapper = LinearColorMapper(palette=PuBu[8][::-1][2:],
                                     low=-4, high=1)
    color_mapper.low_color = 'white'
    color_mapper.high_color = 'black'

    orders = 'Orders 1 & 2' if instrument == 'NIRCam' else 'Order 1'
    s2 = figure(
        tools=TOOLS, width=500, height=500, title='{} {} Contamination with {}'.format(
            orders, targetName, instrument), x_range=Range1d(
            xlim0, xlim1), y_range=Range1d(
                ylim0, ylim1))

    contamO1 = contamO1 if 'NIRCam' in instrument else contamO1.T
    contamO1 = np.fliplr(contamO1) if (
        instrument == 'MIRI') or (
        instrument == 'NIRCam F322W2') else contamO1
    fig_data = np.log10(np.clip(contamO1, 1.e-10, 1.)
                        )  # [:, :361] # might this
    # index have somethig to
    # do w the choppiness
    # of o1 in all instruments

    X = xlim1 if (instrument == 'MIRI') or (
        instrument == 'NIRCam F322W2') else xlim0
    DW = xlim0 - xlim1 if (instrument == 'MIRI') or (instrument ==
                                                     'NIRCam F322W2') else xlim1 - xlim0

    # Begin plotting ~~~~~~~~~~~~~~~~~~~~~~~~

    s2.image([fig_data], x=xlim0, y=ylim0, dw=xlim1 - xlim0, dh=ylim1 - ylim0,
             color_mapper=color_mapper)
    s2.xaxis.axis_label = 'Wavelength (um)'
    if instrument != 'NIRISS':
        s2.yaxis.axis_label = 'Aperture Position Angle (degrees)'

    # Add bad PAs
    bad_PA_color = '#555555'
    bad_PA_alpha = 0.6
    # for ybad0, ybad1 in badPA:
    if len(badPAs) > 0:

        tops, bottoms, lefts, rights = [], [], [], []
        for idx in range(0, len(badPAs)):
            PAgroup = badPAs[idx]
            top_idx = np.max(PAgroup)
            bot_idx = np.min(PAgroup)

            tops.append(top_idx)
            bottoms.append(bot_idx)
            lefts.append(xlim0)
            rights.append(xlim1)

        s2.quad(top=tops, bottom=bottoms,
                left=lefts, right=rights,
                color=bad_PA_color, alpha=bad_PA_alpha)

    # Line plot
    #ax = 1 if 'NIRCam' in instrument else 0
    channels = cols if 'NIRCam' in instrument else rows
    s3 = figure(tools=TOOLS, width=150, height=500,
                x_range=Range1d(0, 100), y_range=s2.y_range, title=None)

    try:
        s3.line(100 * np.sum(contamO1 >= 0.001, axis=1) / channels, PA - dPA / 2,
                line_color='blue', legend_label='> 0.001')
        s3.line(100 * np.sum(contamO1 >= 0.01, axis=1) / channels, PA - dPA / 2,
                line_color='green', legend_label='> 0.01')
    except AttributeError:
        s3.line(100 * np.sum(contamO1 >= 0.001, axis=1) / channels, PA - dPA / 2,
                line_color='blue', legend='> 0.001')
        s3.line(100 * np.sum(contamO1 >= 0.01, axis=1) / channels, PA - dPA / 2,
                line_color='green', legend='> 0.01')

    s3.xaxis.axis_label = '% channels contam.'
    s3.yaxis.major_label_text_font_size = '0pt'

    # ~~~~~~ Order 2 ~~~~~~
    # Contam plot
    if instrument == 'NIRISS':
        xlim0 = lamO2.min()
        xlim1 = lamO2.max()
        ylim0 = PA.min() - 0.5 * dPA
        ylim1 = PA.max() + 0.5 * dPA
        xlim0 = 0.614
        s5 = figure(
            tools=TOOLS,
            width=500,
            height=500,
            title='Order 2 {} Contamination with {}'.format(
                targetName,
                instrument),
            x_range=Range1d(
                xlim0,
                xlim1),
            y_range=s2.y_range)
        fig_data = np.log10(np.clip(contamO2.T, 1.e-10, 1.))[:, 300:]
        s5.image(
            [fig_data],
            x=xlim0,
            y=ylim0,
            dw=xlim1 - xlim0,
            dh=ylim1 - ylim0,
            color_mapper=color_mapper)
        #s5.yaxis.major_label_text_font_size = '0pt'
        s5.xaxis.axis_label = 'Wavelength (um)'
        s5.yaxis.axis_label = 'Aperture Position Angle (degrees)'

        if len(badPAs) > 0:

            tops, bottoms, lefts, rights = [], [], [], []
            for idx in range(0, len(badPAs)):
                PAgroup = badPAs[idx]
                top_idx = np.max(PAgroup)
                bot_idx = np.min(PAgroup)

                tops.append(top_idx)
                bottoms.append(bot_idx)
                lefts.append(xlim0)
                rights.append(xlim1)

            s5.quad(top=tops, bottom=bottoms,
                    left=lefts, right=rights,
                    color=bad_PA_color, alpha=bad_PA_alpha)

        # Line plot
        s6 = figure(tools=TOOLS, width=150, height=500, y_range=s2.y_range,
                    x_range=Range1d(100, 0), title=None)

        try:
            s6.line(100 * np.sum(contamO2 >= 0.001, axis=0) / rows, PA - dPA / 2,
                    line_color='blue', legend_label='> 0.001')
            s6.line(100 * np.sum(contamO2 >= 0.01, axis=0) / rows, PA - dPA / 2,
                    line_color='green', legend_label='> 0.01')
        except AttributeError:
            s6.line(100 * np.sum(contamO2 >= 0.001, axis=0) / rows, PA - dPA / 2,
                    line_color='blue', legend='> 0.001')
            s6.line(100 * np.sum(contamO2 >= 0.01, axis=0) / rows, PA - dPA / 2,
                    line_color='green', legend='> 0.01')

        s6.xaxis.axis_label = '% channels contam.'
        s6.yaxis.major_label_text_font_size = '0pt'

    if len(badPAs) > 0:

        tops, bottoms, lefts, rights = [], [], [], []
        for idx in range(0, len(badPAs)):
            PAgroup = badPAs[idx]
            top_idx = np.max(PAgroup)
            bot_idx = np.min(PAgroup)

            tops.append(top_idx)
            bottoms.append(bot_idx)
            lefts.append(0)
            rights.append(100)

        s3.quad(top=tops, bottom=bottoms,
                left=lefts, right=rights,
                color=bad_PA_color, alpha=bad_PA_alpha)
        if instrument == 'NIRISS':
            s6.quad(top=tops, bottom=bottoms,
                    left=rights, right=lefts,
                    color=bad_PA_color, alpha=bad_PA_alpha)

    # ~~~~~~ Plotting ~~~~~~

    if instrument != 'NIRISS':
        fig = gridplot(children=[[s2, s3]])
    else:
        fig = gridplot(children=[[s6, s5, s2, s3]])

    return fig  # , contamO1


if __name__ == "__main__":
    # arguments RA & DEC, conversion to radians
    argv = sys.argv

    ra = argv[1]
    dec = argv[2]
    cubeNameSuf = argv[3]

    pamin = 0 if len(argv) < 5 else int(argv[4])
    pamax = 360 if len(argv) < 6 else int(argv[5])

    cubeName = argv[6]
    targetName = None if len(argv) < 8 else argv[7]
    save = False if len(argv) < 8 else True  # if name provided -> save
    tmpDir = "." if len(argv) < 9 else argv[8]
    os.makedirs(tmpDir, exist_ok=True)

    goodPA, badPA, _ = vpa.checkVisPA(ra, dec, targetName)

    contam(cubeName, targetName=targetName, paRange=[pamin, pamax],
           badPA=badPA, tmpDir=tmpDir)



"""Base and child classes to handle light curve fitting

Author: Joe Filippazzo
Email: jfilippazzo@stsci.edu
"""
import numpy as np
import pandas as pd
from bokeh.plotting import figure, show

from .models import Model, CompositeModel
from .fitters import lmfitter
from ..utils import COLORS


class LightCurveFitter:
    def __init__(self, time, flux, model):
        """Fit the model to the flux cube

        Parameters
        ----------
        time:
            1D or 2D time axes
        flux:
            2D flux
        """
        self.flux = np.ones(100)
        self.time = np.arange(100)
        self.results = pd.DataFrame(names=('fit_number', 'wavelength', 'P',
                                           'Tc', 'a/Rs', 'b', 'd', 'ldcs',
                                           'e', 'w', 'model_name', 'chi2'))

    def run(self):
        """Run the model fits"""
        pass

    # Method to return sliced results table
    def master_slicer(self, value, param_name='wavelength'):
        return self.results.iloc[self.results[param_name] == value]


class LightCurve(Model):
    def __init__(self, time, flux, unc=None, parameters=None, units='MJD', name='My Light Curve'):
        """
        A class to store the actual light curve

        Parameters
        ----------
        time: sequence
            The time axis in days, [MJD or BJD]
        flux: sequence
            The flux in electrons (not ADU)
        unc: sequence
            The uncertainty on the flux
        parameters: str, object (optional)
            The orbital parameters of the star/planet system,
            may be a path to a JSON file or a parameter object
        units: str
            The time units
        name: str
            A name for the object
        """
        # Initialize the model
        super().__init__()

        # Check data
        if len(time) != len(flux):
            raise ValueError('Time and flux axes must be the same length.')

        # Set the data arrays
        if unc is not None:
            if len(unc) != len(time):
                raise ValueError('Time and unc axes must be the same length.')

            self.unc = unc

        else:
            self.unc = np.array([np.nan]*len(self.time))

        # Set the time and flux axes
        self.time = time
        self.flux = flux

        # Set the units
        self.units = units
        self.name = name

        # Place to save the fit results
        self.results = []

    def fit(self, model, fitter='lmfit', **kwargs):
        """Fit the model to the lightcurve

        Parameters
        ----------
        model: ExoCTK.lightcurve_fitter.models.Model
            The model to fit to the data
        fitter: str
            The name of the fitter to use
        """
        # Empty default fit
        fit_model = None

        # Make sure the model is a CompositeModel
        if not isinstance(model, type(CompositeModel)):
            model = CompositeModel([model])

        if fitter == 'lmfit':

            # Run the fit
            fit_model = lmfitter(self.time, self.flux, model, self.unc, **kwargs)

        else:
            raise ValueError("{} is not a valid fitter.".format(fitter))

        # Store it
        if fit_model is not None:
            self.results.append(fit_model)

    def plot(self, fits=True, draw=True):
        """Plot the light curve with all available fits

        Parameters
        ----------
        fits: bool
            Plot the fit models
        draw: bool
            Show the figure, else return it

        Returns
        -------
        bokeh.plotting.figure
            The figure
        """
        # Make the figure
        fig = figure(width=800, height=400)

        # Draw the data
        fig.circle(self.time, self.flux, legend=self.name)

        # Plot fit models
        if fits and len(self.results) > 0:
            for model in self.results:
                model.plot(self.time, fig=fig, color=next(COLORS))

        # Format axes
        fig.xaxis.axis_label = str(self.units)
        fig.yaxis.axis_label = 'Flux'

        # Draw or return
        if draw:
            show(fig)
        else:
            return fig

    def reset(self):
        """Reset the results"""
        self.results = []



""" The contamVerify mini tool will be a companion to ExoCTK's Contamination
Overlap tool, as it will visualize the Contaminaton Bokeh plots on the website.

Functions are:
plotTemps    - Plots the temperatures of stars according to color.
traceLengths - Fine-tunes the trace lengths in the plot.
contamVerify - The main mini tool. Outputs a .pdf file with one or more figures
               showing the FOV in the science frame according to the input
               Aperture Position Angle(s) it is fed.

Author(s)
---------
Jennifer V. Medina, 2020
"""
import astropy.coordinates as crd
import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import os
import pysiaf

from astropy.io import fits
from astroquery.irsa import Irsa
from matplotlib import cm
from matplotlib.backends.backend_pdf import FigureCanvasPdf, PdfPages
from matplotlib.figure import Figure
from scipy.io import readsav

EXOCTK_DATA = os.environ.get('EXOCTK_DATA')
if not EXOCTK_DATA:
    print(
        'WARNING: The $EXOCTK_DATA environment variable is not set. Contamination overlap will not work. Please set the '
        'value of this variable to point to the location of the exoctk_data '
        'download folder.  Users may retreive this folder by clicking the '
        '"ExoCTK Data Download" button on the ExoCTK website, or by using '
        'the exoctk.utils.download_exoctk_data() function.')
    TRACES_PATH = None

TRACES_PATH = os.path.join(EXOCTK_DATA, 'exoctk_contam', 'traces')


def plotTemps(TEMPS, allRA, allDEC):
    """ The stars' colors in the plot will be a function of effective stellar
    temperatures when plotting with this function. """

    # Getting the color palette
    colors = cm.get_cmap('viridis', len(TEMPS))
    colors_0 = np.asarray(colors.colors)

    # Assigning index arrays to TEMPS array
    i = TEMPS.argsort()
    ii = TEMPS.argsort().argsort()

    # Matching the colors to the corresponding magnitude
    colors = colors_0
    starsx, starsy = allRA[i], allDEC[i]

    plt.style.use('dark_background')
    for x, y, c in zip(starsx, starsy, colors):
        plt.scatter(
            x,
            y,
            marker='*',
            s=150,
            color=c,
            picker=True,
            lw=0.5,
            edgecolor='white')

    # Colorbar
    sm = plt.cm.ScalarMappable(cmap=plt.cm.viridis,
                               norm=plt.Normalize(vmin=TEMPS.min(),
                                                  vmax=TEMPS.max()))
    sm._A = []

    cbar = plt.colorbar(sm, fraction=0.046, pad=0.04)
    cbar.set_label('effective Temperature (K)', fontsize=20)


def traceLength(inst):
    """ For fine-tuning the trace lengths in the contamVerify output figures """

    # Getting example trace to calculate rough estimate of trace lengths
    if 'NIRCam' in inst:
        FILE = 'rot_o1_6000.0.fits'
    elif 'MIRI' in inst:
        FILE = 'LOWbg_6000.0.fits'
    elif 'NIRISS' in inst:
        FILE = 'modelOrder12_teff6000.sav'

    trFile = os.path.join(TRACES_PATH, inst.replace(' ', '_'), FILE)
    trData = readsav(trFile)['modelo12'] if 'NIRISS' in inst \
        else fits.getdata(trFile, 1)
    trData = trData[0]
    print(np.shape(trData))
    ax = 1 if 'NIRCam' in inst else 0
    peak = trData.max()

    # the length of the trace
    targ_trace_start = np.where(trData > 0.0001 * peak)[ax].min()
    targ_trace_stop = np.where(trData > 0.0001 * peak)[ax].max()

    return targ_trace_start, targ_trace_stop


def contamVerify(RA, DEC, INSTRUMENT, APAlist, binComp=[], PDF='', web=False):
    """ Generates a PDF file of figures displaying a simulation
    of the science image for any given observation using the parameters provided.

    Parameter(s)
    ------------
    RA  : str
        The Right Ascension of your target in HH:MM:SS
    DEC : str
        The Declination of your target in DD:MM:SS
    INSTRUMENT : str
        The instrument you are observing with (case-sensitive).
        The software currently supports:
        'MIRI', 'NIRISS', 'NIRCam F322W2', 'NIRCam F444W'
    APAlist : list
        A list of Aperture Position Angle(s). Element(s) must be in integers.
        Example 1:
        [1, 25, 181, 205]
        Example 2:
        [25]
    binComp : list
        A list containing parameters of a missing companion that is not
        in the 2MASS IRSA point-source catalog. The format is:
        [RA (arcseconds), DEC (arcseconds), J mag, H mag, K mag]
        [string, string, integer, integer, integer]
    PDF : string
        The path to where the PDF file will be saved. If left blank, the PDF
        file will be saved in your current working directory.
        Example:
        'path/to/my/file.pdf'
    web : boolean
        Makes it easier to integrate it onto the website. Leave this as false,
        unless you're running this in app_exoctk.py

    Returns
    -------
    A .PDF file containing a simulation of the FOV of your target in the
    science coordinate system. Some things to consider when reading the figures:

    1. The target is circled in red
    2. Stellar temperatures of all sources are plotted by color
    3. The gray region oulined in blue represents the aperture for the given
       instrument.
    4. The blue square represents the readout region, or the "origin"

    """
    print('Generating FOV...')
    # Decimal degrees --> HMSDMS for Irsa.query_region()
    targetcrd = crd.SkyCoord(ra=RA, dec=DEC, unit='deg').to_string('hmsdms')
    targetRA, targetDEC = RA, DEC

    # Querying for neighbors with 2MASS IRSA's fp_psc (point-source catalog)
    rad = 2.5
    print('Querying for point-sources within {} arcminutes...'.format(str(rad)))
    info = Irsa.query_region(
        targetcrd,
        catalog='fp_psc',
        spatial='Cone',
        radius=rad * u.arcmin)

    # Coordinates of all stars in FOV, including target
    allRA = info['ra'].data.data
    allDEC = info['dec'].data.data

    # Initiating a dictionary to hold all relevant star information
    stars = {}
    stars['RA'], stars['DEC'] = allRA, allDEC

    print('Total point-sources found in region: {}'.format(len(stars['RA'])))
    # Finding the target using relative distances
    sindRA = (targetRA - stars['RA']) * np.cos(targetDEC)
    cosdRA = targetDEC - stars['DEC']
    distance = np.sqrt(sindRA**2 + cosdRA**2)
    targetIndex = np.argmin(distance)

    # Appending missing companion to the above lists (if any)
    if binComp != []:
        print('Adding missing companion...')
        bb = binComp[0] / 3600 / np.cos(allDEC[targetIndex] * deg2rad)
        allRA = np.append(allRA, (allRA[targetIndex] + bb))
        allDEC = np.append(allDEC, (allDEC[targetIndex] + binComp[1] / 3600))
        Jmag = np.append(Jmag, binComp[2])
        Hmag = np.append(Kmag, binComp[3])
        Kmag = np.append(Kmag, binComp[4])
        J_Hobs = Jmag - Hmag
        H_Kobs = Hmag - Kmag

    # Restoring model parameters
    modelParam = readsav(os.path.join(TRACES_PATH, 'NIRISS', 'modelsInfo.sav'),
                         verbose=False)
    models = modelParam['models']
    modelPadX = modelParam['modelpadx']
    modelPadY = modelParam['modelpady']
    dimXmod = modelParam['dimxmod']
    dimYmod = modelParam['dimymod']
    jhMod = modelParam['jhmod']
    hkMod = modelParam['hkmod']
    teffMod = modelParam['teffmod']

    # JHK bands of all stars in FOV, including target
    Jmag = info['j_m'].data.data
    Hmag = info['h_m'].data.data
    Kmag = info['k_m'].data.data
    # J-H band, H-K band. This will be used to derive the stellar Temps later
    J_Hobs = Jmag - Hmag
    H_Kobs = Hmag - Kmag

    # Number of stars
    nStars = stars['RA'].size

    # Find/assign Teff of each star
    print('Calculating effective temperatures...')
    starsT = np.empty(nStars)
    for j in range(nStars):
        color_separation = (J_Hobs[j] - jhMod)**2 + (H_Kobs[j] - hkMod)**2
        min_separation_ind = np.argmin(color_separation)
        starsT[j] = teffMod[min_separation_ind]

    # Record keeping
    stars['Temp'] = starsT

    # Initiating a dictionary for customizability
    apertures = {}
    apertures['NIRISS'] = ['NIS_SOSSFULL', 'NIS_SOSSFULL']
    apertures['NIRCam F444W'] = ['NRCA5_GRISM256_F444W', 'NRCA5_FULL']
    apertures['NIRCam F322W2'] = ['NRCA5_GRISM256_F322W2', 'NRCA5_FULL']
    apertures['MIRI'] = ['MIRIM_SLITLESSPRISM', 'MIRIM_FULL']

    # Instantiate SIAF object
    siaf = pysiaf.Siaf(INSTRUMENT.split(' ')[0])

    aper = siaf.apertures[apertures[INSTRUMENT][0]]
    full = siaf.apertures[apertures[INSTRUMENT][1]]

    # DET_targ -> TEL_targ -> get attitude matrix for target
    # -> TEL_neighbor -> DET_neighbor -> SCI_neighbor
    print('Converting Sky --> Science coordinates...')
    xSweet, ySweet = aper.reference_point('det')

    v2targ, v3targ = aper.det_to_tel(xSweet, ySweet)

    contam = {}

    if not web:
        filename = 'contam_{}_{}_{}.pdf'.format(RA, DEC, INSTRUMENT)
        defaultPDF = os.path.join(os.getcwd(), filename).replace(' ', '_')
        PDF = defaultPDF if PDF == '' else PDF
    elif web:
        filename = 'contam_{}_{}_{}.pdf'.format(RA, DEC, INSTRUMENT)
        PDF = os.path.join(TRACES_PATH, filename)

    print('Saving figures to: {}'.format(PDF))
    print('This will take a second...')
    pdfobj = PdfPages(PDF)
    for APA in APAlist:

        attitude = pysiaf.utils.rotations.attitude_matrix(
            v2targ, v3targ, targetRA, targetDEC, APA)

        xdet, ydet = [], []
        xsci, ysci = [], []

        for starRA, starDEC in zip(stars['RA'], stars['DEC']):
            # Get the TEL coordinates of each star using the attitude
            # matrix of the target
            V2, V3 = pysiaf.utils.rotations.sky_to_tel(
                attitude, starRA, starDEC)
            # Convert to arcsec and turn to a float
            V2, V3 = V2.to(u.arcsec).value, V3.to(u.arcsec).value

            XDET, YDET = aper.tel_to_det(V2, V3)
            XSCI, YSCI = aper.det_to_sci(XDET, YDET)

            xdet.append(XDET)
            ydet.append(YDET)
            xsci.append(XSCI)
            ysci.append(YSCI)

        XDET, YDET = np.array(xdet), np.array(ydet)
        XSCI, YSCI = np.array(xsci), np.array(ysci)

        starsAPA = {'xdet': XDET, 'ydet': YDET, 'xsci': XSCI, 'ysci': YSCI}

        # Finding indexes of neighbor sources that land on detector
        rows, cols = full.corners('det')

        minrow, maxrow = rows.min(), rows.max()
        mincol, maxcol = cols.min(), cols.max()

        inFOV = []
        for star in range(0, nStars):

            x, y = starsAPA['xdet'][star], starsAPA['ydet'][star]
            if (mincol < x) & (x < maxcol) & (minrow < y) & (y < maxrow):
                inFOV.append(star)

        inFOV = np.array(inFOV)

        # Making final plot
        fig = plt.figure(figsize=(15, 15))
        aper.plot(frame='sci', fill_color='gray', color='blue')
        plt.scatter(
            XSCI[targetIndex],
            YSCI[targetIndex],
            s=400,
            lw=1.5,
            facecolor='gray',
            edgecolor='red')
        plotTemps(starsT[inFOV], XSCI[inFOV], YSCI[inFOV])
        aper.plot_frame_origin(frame='sci', which='sci')

        # Fine-tune trace lengths
        start, stop = traceLength(INSTRUMENT)

        # Plotting the trace footprints
        for x, y in zip(XSCI[inFOV], YSCI[inFOV]):

            if 'F322W2' in INSTRUMENT:
                plt.plot([x - stop, x + start], [y, y],
                         lw=40, color='white', alpha=0.2)
                plt.plot([x - stop, x + start],
                         [y, y], lw=2., color='white')
            elif 'F444W' in INSTRUMENT:
                plt.plot([x - start, x + stop], [y, y],
                         lw=40, color='white', alpha=0.2)
                plt.plot([x - start, x + stop],
                         [y, y], lw=2., color='white')
            else:
                plt.plot([x, x], [y - stop, y + start],
                         lw=40, color='white', alpha=0.2)
                plt.plot([x, x], [y - stop, y + start],
                         lw=2., color='white')

        # Labeling
        aperstr = str(aper.AperName.replace('_', ' '))
        tx, ty = str(
            round(
                XSCI[targetIndex])), str(
            round(
                YSCI[targetIndex]))
        plt.title(
            'The FOV in SCIENCE coordinates at APA {}$^o$'.format(
                str(APA)) +
            '\n' +
            '{}'.format(aperstr) +
            '\n' +
            'Target (X,Y): {}, {}'.format(
                tx,
                ty),
            fontsize=20)

        # Adding to PDF
        pdfobj.savefig(fig, bbox_inches='tight')

    pdfobj.close()

    if web:
        return PDF



#!/usr/bin/env python

"""This module contains various functions for interacting with AWS for
``exoctk`` atmospheric retrievals.

Authors
-------

    - Matthew Bourque

Use
---

    This script is inteneded to be imported and used by other modules,
    for example:

        from aws_tools import get_config
        get_config()

Dependencies
------------

    Dependent libraries include:

    - boto3
    - paramiko
    - scp

    Users must also have a ``aws_config.json`` file present within the
    ``atmospheric_retrievals`` subdirectory.  This file must be of
    a valid JSON format and contain two key/value pairs,
    ``ec2_id`` and ``ssh_file``, e.g.:

    {
    "ec2_id" : "lt-021de8b904bc2b728",
    "ssh_file" : "~/.ssh/my_ssh_key.pem"
    }

    where the ``ec2_id`` contains the ID for an EC2 launch template
    or an existing EC2 instance, and ``ssh_file`` points to the SSH
    public key used for logging into an AWS account.
"""

import json
import logging
import os
import time

import boto3
import paramiko
from scp import SCPClient


def build_environment(instance, key, client):
    """Builds an ``exoctk`` environment on the given AWS EC2 instance

    Parameters
    ----------
    instance : obj
        A ``boto3`` AWS EC2 instance object.
    key : obj
        A ``paramiko.rsakey.RSAKey`` object.
    client : obj
        A ``paramiko.client.SSHClient`` object.
    """

    logging.info('Building ExoCTK environment')

    # Connect to the EC2 instance and run commands
    connected = False
    iterations = 0
    while not connected:
        if iterations == 12:
            logging.critical('Could not connect to {}'.format(instance.public_dns_name))
            break
        try:
            client.connect(hostname=instance.public_dns_name, username='ec2-user', pkey=key)
            scp = SCPClient(client.get_transport())
            scp.put('build-exoctk-env-cpu.sh', '~/build-exoctk-env-cpu.sh')
            stdin, stdout, stderr = client.exec_command('chmod 700 build-exoctk-env-cpu.sh && ./build-exoctk-env-cpu.sh')
            connected = True
        except:
            iterations += 1
            time.sleep(5)

    output = stdout.read()
    log_output(output)


def get_config():
    """Return a dictionary that holds the contents of the
    ``aws_config.json`` config file.

    Returns
    -------
    settings : dict
        A dictionary that holds the contents of the config file.
    """

    config_file_location = os.path.join(os.path.dirname(__file__), 'aws_config.json')

    if not os.path.isfile(config_file_location):
        raise FileNotFoundError('Missing AWS configuration file ("aws_config.json")')

    with open(config_file_location, 'r') as config_file:
        settings = json.load(config_file)

    return settings


def log_output(output):
    """Logs the given output of the EC2 instance.

    Parameters
    ----------
    output : str
        The standard output of the EC2 instance
    """

    output = output.decode("utf-8")
    output = output.replace('\t', '  ').replace('\r', '').replace("\'", "").split('\n')
    for line in output:
        logging.info(line)


def start_ec2(ssh_file, ec2_id):
    """Create a new EC2 instance or start an existing EC2 instance.

    A new EC2 instance will be created if the supplied ``ec2_id`` is an
    EC2 template ID.  An existing EC2 instance will be started if the
    supplied ``ec2_id`` is an ID for an existing EC2 instance.

    Parameters
    ----------
    ssh_file : str
        Relative path to SSH public key to be used by AWS (e.g.
        ``~/.ssh/exoctk.pem``).
    ec2_id : str
        The AWS EC2 template id (e.g. ``lt-021de8b904bc2b728``) or
        instance ID (e.g. ``i-0d0c8ca4ab324b260``).

    Returns
    -------
    instance : obj
        A ``boto3`` AWS EC2 instance object.
    key : obj
        A ``paramiko.rsakey.RSAKey`` object.
    client : obj
        A ``paramiko.client.SSHClient`` object.
    """

    ec2 = boto3.resource('ec2')

    # If the given ec2_id is for an EC2 template, then create the EC2 instance
    if ec2_id.split('-')[0] == 'lt':
        LaunchTemplate = {'LaunchTemplateId': ec2_id}
        instances = ec2.create_instances(
            LaunchTemplate=LaunchTemplate,
            MaxCount=1,
            MinCount=1)
        instance = instances[0]
        logging.info('Launched EC2 instance {}'.format(instance.id))

    # If the given ec2_id is for an existing EC2 instance, then start it
    else:
        instance = ec2.Instance(ec2_id)
        instance.start()
        logging.info('Started EC2 instance {}'.format(ec2_id))

    instance.wait_until_running()
    instance.load()

    # Establish SSH key and client
    key = paramiko.RSAKey.from_private_key_file(ssh_file)
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    return instance, key, client


def stop_ec2(ec2_id, instance):
    """Terminates or stops the given AWS EC2 instance.

    The instance is terminated if the supplied ``ec2_id`` is a EC2
    template ID.  The instance is stopped if the supplied ``ec2_id``
    is an ID for a particular EC2 instance.

    Parameters
    ----------
    ec2_id : str
        The AWS EC2 template id (e.g. ``lt-021de8b904bc2b728``) or
        instance ID (e.g. ``i-0d0c8ca4ab324b260``).
    instance : obj
        A ``boto3`` AWS EC2 instance object.
    """

    ec2 = boto3.resource('ec2')

    # If the given ec2_id is for an EC2 template, then terminate the EC2 instance
    if ec2_id.split('-')[0] == 'lt':
        ec2.instances.filter(InstanceIds=[instance.id]).terminate()
        logging.info('Terminated EC2 instance {}'.format(instance.id))

    # If the given ec2_id is for an existing EC2 instance, then stop it
    else:
        instance.stop()
        logging.info('Stopped EC2 instance {}'.format(ec2_id))


def transfer_from_ec2(instance, key, client, filename):
    """Copy files from EC2 user back to the user

    Parameters
    ----------
    instance : obj
        A ``boto3`` AWS EC2 instance object.
    key : obj
        A ``paramiko.rsakey.RSAKey`` object.
    client : obj
        A ``paramiko.client.SSHClient`` object.
    filename : str
        The path to the file to transfer
    """

    logging.info('Copying {} from EC2'.format(filename))

    client.connect(hostname=instance.public_dns_name, username='ec2-user', pkey=key)
    scp = SCPClient(client.get_transport())
    scp.get(filename)


def transfer_to_ec2(instance, key, client, filename):
    """Copy parameter file from user to EC2 instance

    Parameters
    ----------
    instance : obj
        A ``boto3`` AWS EC2 instance object.
    key : obj
        A ``paramiko.rsakey.RSAKey`` object.
    client : obj
        A ``paramiko.client.SSHClient`` object.
    filename : str
        The path to the file to transfer
    """

    logging.info('Copying {} to EC2'.format(filename))

    connected = False
    iterations = 0
    while not connected:
        if iterations >= 10:
            logging.critical('Could not connect to {}'.format(instance.public_dns_name))
            break
        try:
            client.connect(hostname=instance.public_dns_name, username='ec2-user', pkey=key)
            scp = SCPClient(client.get_transport())
            scp.put(filename)
            connected = True
        except:
            logging.warning('Could not connect to {}, retrying.'.format(instance.public_dns_name))
            time.sleep(5)
            iterations += 1



"""
Package to calculate target visibility and contamination
"""
from . import astro_funcx
from . import math_extensionsx
from . import ephemeris_old2x
from . import visibilityPA
from . import contamination_figure
from . import resolve
from . import time_extensionsx
from . import f_visibilityPeriods
from . import quaternionx
from . import field_simulator



#!/usr/bin/python
# -*- coding: latin-1 -*-
"""
A module for managing references in ExoCTK
"""
import os
import pkg_resources

import bibtexparser as bt


class References(object):
    """
    Creates and manages a References object to track references
    within an ExoCTK user session

    Attributes
    ----------
    bibfile: str
        The path to the bibtex file from which the references will be read
    refs: list
        The list of bibcodes saved during the user session
    database: bibtexparser.bibdatabase.BibDatabase object
        The database of parsed bibtex entries
    bibcodes: list
        The list of all bibcodes in the database

    """
    def __init__(self, bibfile=''):
        """
        Initializes an empty References object which points to a
        .bib file

        Parameters
        ----------
        bibfile: str
          The path to the bibtex file from which the references will be read

        """
        bibfile = bibfile or \
            pkg_resources.resource_filename('ExoCTK', 'data/core/bibtex.bib')

        # Attributes for the filepath and references
        self.bibfile = bibfile
        self.refs = []

        # Load the bibtex into a database
        bf = open(bibfile)
        self.database = bt.load(bf)
        bf.close()

        # The list of all bibcodes in the bibfile
        self.bibcodes = [i['ID'] for i in self.database.entries]

    def add(self, bibcode):
        """
        Adds a bibcode to the References object

        Parameters
        ----------
        bibcode: str
            The unique compact identifier for the reference to be added

        """
        # Check that the bibcode is in the bibtex file
        if bibcode in self.bibcodes:
            self.refs += [bibcode]
            print(bibcode, 'added to list of references.')

        # Suggest adding it to the bibfile
        else:
            print(bibcode, 'not in bibfile at', self.bibfile)
            print('Add the bibtex entry to the file and try agin.')

    def remove(self, bibcode):
        """
        Removes a bibcode from the References object

        Parameters
        ----------
        bibcode: str
            The unique compact identifier for the reference to be removed

        """
        # Check that the bibcode is in the bibtex file
        if bibcode in self.bibcodes:
            self.refs = [r for r in self.refs if r != bibcode]
            print(bibcode, 'removed from list of references.')

        # Nothing to remove!
        else:
            print(bibcode, 'not in bibfile at', self.bibfile)

    def write(self, filepath=''):
        """
        Write the .bib file

        Parameters
        ----------
        filepath: str
            The directory to which the .bib file should be written.
            Can be an existing .bib file or just a directory

        """
        # Use existing .bib file or create new one
        if filepath.endswith('.bib'):
            bibfile = filepath
        else:
            bibfile = os.path.join(filepath, 'biblio.bib')

        # Create a new database instance
        final = bt.bibdatabase.BibDatabase()

        # Get the relevant bibtex entries
        final.entries = [d for d in self.database.entries
                         if d['ID'] in list(set(self.refs))]

        # Write the bibtex to file
        with open(bibfile, 'w') as out:
            out.write(bt.bwriter.BibTexWriter().write(final))



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

REQUIRES = ['asteval',
            'astropy',
            'astroquery',
            'bandit',
            'bibtexparser',
            'bokeh',
            'boto3',
            'corner',
            'cython',
            'docopt',
            'docutils==0.14',
            'flake8',
            'flask',
            'flask_wtf',
            'gunicorn',
            'h5py',
            'ipython',
            'lmfit',
            'matplotlib',
            'numpy',
            'numpydoc',
            'pandas',
            'paramiko',
            'platon',
            'pysiaf',
            'pysynphot',
            'pytest',
            'pyyaml==5.1.0',
            'scipy',
            'scp',
            'sphinx',
            'sphinx_astropy',
            'sqlalchemy',
            'svo_filters',
            'wtforms',
            'werkzeug==0.16.1']

DEPENDENCY_LINKS = ['git+https://github.com/spacetelescope/jwst_gtvt.git@cd6bc76f66f478eafbcc71834d3e735c73e03ed5']

FILES = []
for root, _, files in os.walk("exoctk"):
    FILES += [os.path.join(root.replace("exoctk/", ""), fname) \
        for fname in files if not fname.endswith(".py") and not fname.endswith(".pyc")]

setup(name='exoctk',
      version='1.1.0',
      description='Observation reduction and planning tools for exoplanet science',
      packages=find_packages(".", exclude=["*.tests"]),
      package_data={'exoctk': FILES},
      install_requires=REQUIRES,
      dependency_links=DEPENDENCY_LINKS,
      author='The ExoCTK Group',
      author_email='exoctk@gmail.com',
      license='MIT',
      url='https://github.com/ExoCTK/exoctk',
      long_description='',
      zip_safe=True,
      use_2to3=False
)



"""Base and child classes to handle models
used to fit light curves

Author: Joe Filippazzo
Email: jfilippazzo@stsci.edu
"""
import copy
import inspect
import os

import astropy.units as q
try:
    import batman
except ImportError:
    print("Could not import batman. Functionality may be limited.")
from bokeh.plotting import figure, show
import numpy as np

from .parameters import Parameters
from ..utils import COLORS
from ..limb_darkening.limb_darkening_fit import ld_profile


class Model:
    def __init__(self, **kwargs):
        """
        Create a model instance
        """
        # Set up model attributes
        self.name = 'New Model'
        self._time = None
        self._flux = None
        self._units = q.day
        self._parameters = Parameters()
        self.components = None
        self.fmt = None

        # Store the arguments as attributes
        for arg, val in kwargs.items():
            setattr(self, arg, val)

    def __mul__(self, other):
        """Multiply model components to make a combined model

        Parameters
        ----------
        other: ExoCTK.lightcurve_fitting.models.Model
            The model to multiply

        Returns
        -------
        ExoCTK.lightcurve_fitting.lightcurve.Model
            The combined model
        """
        # Make sure it is the right type
        attrs = ['units', 'flux', 'time']
        if not all([hasattr(other, attr) for attr in attrs]):
            raise TypeError('Only another Model instance may be multiplied.')

        # Combine the model parameters too
        params = self.parameters + other.parameters

        return CompositeModel([copy.copy(self), other], parameters=params)

    @property
    def flux(self):
        """A getter for the flux"""
        return self._flux

    @flux.setter
    def flux(self, flux_array):
        """A setter for the flux

        Parameters
        ----------
        flux_array: sequence
            The flux array
        """
        # Check the type
        if not isinstance(flux_array, (np.ndarray, tuple, list)):
            raise TypeError("flux axis must be a tuple, list, or numpy array.")

        # Set the array
        self._flux = np.array(flux_array)

    def interp(self, new_time):
        """Interpolate the flux to a new time axis

        Parameters
        ----------
        new_time: sequence, astropy.units.quantity.Quantity
            The time array
        """
        # Check the type
        if not isinstance(new_time, (np.ndarray, tuple, list)):
            raise TypeError("Time axis must be a tuple, list, or numpy array")

        # Calculate the new flux
        self.flux = np.interp(new_time, self.time, self.flux)

        # Set the new time axis
        self.time = new_time

    @property
    def parameters(self):
        """A getter for the parameters"""
        return self._parameters

    @parameters.setter
    def parameters(self, params):
        """A setter for the parameters"""
        # Process if it is a parameters file
        if isinstance(params, str) and os.file.exists(params):
            params = Parameters(params)

        # Or a Parameters instance
        if not isinstance(params, (Parameters, type(None))):
            raise TypeError("'params' argument must be a JSON file, ascii\
                             file, or parameters.Parameters instance.")

        # Set the parameters attribute
        self._parameters = params

    def plot(self, time, components=False, fig=None, draw=False, color='blue', **kwargs):
        """Plot the model

        Parameters
        ----------
        time: array-like
            The time axis to use
        components: bool
            Plot all model components
        fig: bokeh.plotting.figure (optional)
            The figure to plot on

        Returns
        -------
        bokeh.plotting.figure
            The figure
        """
        # Make the figure
        if fig is None:
            fig = figure(width=800, height=400)

        # Set the time
        self.time = time

        # Plot the model
        fig.line(self.time, self.eval(**kwargs), legend=self.name, color=color)

        if components and self.components is not None:
            for comp in self.components:
                fig = comp.plot(self.time, fig=fig, draw=False, color=next(COLORS), **kwargs)

        # Format axes
        fig.xaxis.axis_label = str(self.units)
        fig.yaxis.axis_label = 'Flux'

        if draw:
            show(fig)
        else:
            return fig

    @property
    def time(self):
        """A getter for the time"""
        return self._time

    @time.setter
    def time(self, time_array, units='MJD'):
        """A setter for the time

        Parameters
        ----------
        time_array: sequence, astropy.units.quantity.Quantity
            The time array
        units: str
            The units of the input time_array, ['MJD', 'BJD', 'phase']
        """
        # Check the type
        if not isinstance(time_array, (np.ndarray, tuple, list)):
            raise TypeError("Time axis must be a tuple, list, or numpy array.")

        # Set the units
        self.units = units

        # Set the array
        self._time = time_array

    @property
    def units(self):
        """A getter for the units"""
        return self._units

    @units.setter
    def units(self, units):
        """A setter for the units

        Parameters
        ----------
        units: str
            The time units ['BJD', 'MJD', 'phase']
        """
        # Check the type
        if units not in ['BJD', 'MJD', 'phase']:
            raise TypeError("units axis must be 'BJD', 'MJD', or 'phase'.")

        self._units = units


class CompositeModel(Model):
    """A class to create composite models"""
    def __init__(self, models, **kwargs):
        """Initialize the composite model

        Parameters
        ----------
        models: sequence
            The list of models
        """
        # Inherit from Model calss
        super().__init__(**kwargs)

        # Store the models
        self.components = models

    def eval(self, **kwargs):
        """Evaluate the model components"""
        # Get the time
        if self.time is None:
            self.time = kwargs.get('time')

        # Empty flux
        flux = 1.

        # Evaluate flux at each model
        for model in self.components:
            flux *= model.eval(**kwargs)

        return flux


class PolynomialModel(Model):
    """Polynomial Model"""
    def __init__(self, **kwargs):
        """Initialize the polynomial model
        """
        # Inherit from Model class
        super().__init__(**kwargs)

        # Check for Parameters instance
        self.parameters = kwargs.get('parameters')

        # Generate parameters from kwargs if necessary
        if self.parameters is None:
            self._parse_coeffs(kwargs)

    def _parse_coeffs(self, coeff_dict):
        """Convert dict of 'c#' coefficients into a list
        of coefficients in decreasing order, i.e. ['c2','c1','c0']

        Parameters
        ----------
        coeff_dict: dict
            The dictionary of coefficients

        Returns
        -------
        np.ndarray
            The sequence of coefficient values
        """
        params = {cN: coeff for cN, coeff in coeff_dict.items()
                  if cN.startswith('c') and cN[1:].isdigit()}
        self.parameters = Parameters(**params)

        # Parse 'c#' keyword arguments as coefficients
        coeffs = np.zeros(100)
        for k, v in self.parameters.dict.items():
            if k.lower().startswith('c') and k[1:].isdigit():
                coeffs[int(k[1:])] = v[0]

        # Trim zeros and reverse
        self.coeffs = np.trim_zeros(coeffs)[::-1]

    def eval(self, **kwargs):
        """Evaluate the function with the given values"""
        # Get the time
        if self.time is None:
            self.time = kwargs.get('time')

        # Create the polynomial from the coeffs
        poly = np.poly1d(self.coeffs)

        # Convert to local time
        time_local = self.time - self.time.mean()

        # Evaluate the polynomial
        return np.polyval(poly, time_local)


class TransitModel(Model):
    """Transit Model"""
    def __init__(self, **kwargs):
        """Initialize the transit model
        """
        # Inherit from Model calss
        super().__init__(**kwargs)

        # Check for Parameters instance
        self.parameters = kwargs.get('parameters')

        # Generate parameters from kwargs if necessary
        if self.parameters is None:
            self.parameters = Parameters(**kwargs)

        # Store the ld_profile
        self.ld_func = ld_profile(self.parameters.limb_dark.value)
        len_params = len(inspect.signature(self.ld_func).parameters)
        self.coeffs = ['u{}'.format(n) for n in range(len_params)[1:]]

    def eval(self, **kwargs):
        """Evaluate the function with the given values"""
        # Get the time
        if self.time is None:
            self.time = kwargs.get('time')

        # Generate with batman
        bm_params = batman.TransitParams()

        # Set all parameters
        for p in self.parameters.list:
            setattr(bm_params, p[0], p[1])

        # Combine limb darkening coeffs
        bm_params.u = [getattr(self.parameters, u).value for u in self.coeffs]

        # Use batman ld_profile name
        if self.parameters.limb_dark.value == '4-parameter':
            bm_params.limb_dark = 'nonlinear'

        # Make the eclipse
        tt = self.parameters.transittype.value
        m_eclipse = batman.TransitModel(bm_params, self.time, transittype=tt)

        # Evaluate the light curve
        return m_eclipse.light_curve(bm_params)



import numpy as np

from astropy.coordinates import SkyCoord
from bokeh.layouts import gridplot
from bokeh.plotting import figure, show

from exoctk.contam_visibility import visibilityPA as vpa
from exoctk.contam_visibility import field_simulator as fs
from exoctk.contam_visibility import contamination_figure as cf

def main():
    """ Wrapper to the field simulator and contamination figure generator.
    """

    # User inputs
    ra = input('Please input the Right Ascension of your target in decimal degrees (The more decimal places the better) : \n')
    # Making sure RA input is correct before we continue
    if (float(ra) < 0) or (float(ra) > 360):
        print('RA should be between 0 and 360 decimal degrees. Got {}. Starting over...'.format(ra))
        main()

    dec = input('Please input the Declination of your target in decimal degrees (The more decimal places the better) : \n')
    # Making sure DEC input is correct
    if (float(dec) < -90) or (float(dec) > 90):
        print('DEC should be between -90 and +90 in decimal degrees. Got {}. Starting over...'.format(dec))
        main()

    companion = input('Any companion not in IRSA`s 2MASS Point-Source Catalog that should be considered for contamination? If no, just press Enter. If yes, please enter the following (comma-separated, no spaces): \n RA offset ("), DEC offset ("), 2MASS J (mag), H (mag) and Ks (mag) \n ').split(',')
    print(len(companion))
    if len(companion)==5:
        binComp = [int(param) for param in companion]
    elif len(companion)==1:
        binComp = ''
    elif (len(companion) < 5) & (len(companion) > 1):
        print('Companion information is incomplete. Starting over...')
        main()

    instrument = input('Please input the instrument your target will be observed with (Case-sensitive. Can be: NIRISS, MIRI, NIRCam F322W2, or NIRCam F444W) : \n')
    # Making sure the instrument input is correct before we continue
    possible_instruments = ['NIRISS', 'MIRI', 'NIRCam F322W2', 'NIRCam F444W']
    if instrument not in possible_instruments:
        print('It looks like your last input (instrument) had a typo. The instrument input is case-sensitive and can only be NIRISS, MIRI, NIRCam F322W2, or NIRCam F444W. Starting over...')
        main()

    # Getting the bad PAs from visibility calculator for shading purposes
    instrument_vpa = instrument.split(' ')[0]
    paMin, paMax, gd, fig, table, grouped_badPAs = vpa.using_gtvt(ra, dec, instrument_vpa)

    # Converting RA, DEC from decimal degrees (ExoMAST) to HMSDMS
    sc = SkyCoord(ra, dec, unit='deg')
    ra_dec = sc.to_string('hmsdms')
    ra_hms, dec_dms = ra_dec.split(' ')[0], ra_dec.split(' ')[1]

    # Generating cube with a field for every Aperture Position Angle (APA)
    cube = fs.fieldSim(ra_hms, dec_dms, instrument, binComp)

    title_ra, title_dec = str(np.round(float(ra), 3)), str(np.round(float(dec), 3))
    # Generating Bokeh figure `fig` that plots contamination levels at every APA
    plot = cf.contam(cube, instrument, targetName=' {}, {} (RA, DEC)'.format(title_ra, title_dec), badPAs=grouped_badPAs)

    show(plot)

if __name__ == "__main__":
    main()



import numpy as np
import os

from flask_wtf import FlaskForm
from wtforms import StringField, SubmitField, DecimalField, RadioField, SelectField, SelectMultipleField, IntegerField, FloatField
from wtforms.validators import InputRequired, Length, NumberRange, AnyOf, ValidationError
from wtforms.widgets import ListWidget, CheckboxInput

from exoctk.modelgrid import ModelGrid
from exoctk.utils import get_env_variables, FILTERS, PROFILES
from svo_filters import svo


class MultiCheckboxField(SelectMultipleField):
    """Makes a list of checkbox inputs"""
    widget = ListWidget(prefix_label=False)
    option_widget = CheckboxInput()


class BaseForm(FlaskForm):
    """A generic form with target resolve built in"""
    # Target Resolve
    targname = StringField('targname', default='')
    target_url = StringField('target_url', default='')

    # Submit button
    resolve_submit = SubmitField('Resolve Target')


class FortneyModelForm(BaseForm):
    """Form validation for the forward model tools"""
    # Parameters
    planet_teff = SelectField('planet_teff', choices=[(500, '500'), (750, '750'), (1000, '1000'), (1250, '1250'), (1500, '1500'), (1750, '1750'), (2000, '2000'), (2250, '2250'), (2500, '2500')], validators=[InputRequired('An effective temperature is required')])
    planet_mass = DecimalField('planet_mass', default=1.5, validators=[InputRequired('A planet mass is required!'), NumberRange(min=0.0, message='Planet mass must be positive')])
    planet_mass_unit = SelectField('planet_mass_unit', choices=[('M_jup', 'Jupiter Mass'), ('kilogram', 'kilogram'), ('g', 'gram'), ('M_earth', 'Earth Mass'), ('M_sun', 'Solar Mass')], validators=[InputRequired('A mass unit is required')])
    planet_radius = DecimalField('planet_radius', default=1.25, validators=[InputRequired('A planet radius is required!'), NumberRange(min=0, message='Planet radius must be positive')])
    planet_radius_unit = SelectField('planet_radius_unit', choices=[('R_jup', 'Jupiter Radius'), ('kilometer', 'kilometer'), ('m', 'meter'), ('R_earth', 'Earth Radius'), ('R_sun', 'Solar Radius')], validators=[InputRequired('A planet radius unit is required')])
    stellar_radius = DecimalField('stellar_radius', default=1.0, validators=[InputRequired('A stellar radius is required!'), NumberRange(min=0, message='Stellar radius must be positive')])
    stellar_radius_unit = SelectField('stellar_radius_unit', choices=[('R_sun', 'Solar Radius'), ('R_jup', 'Jupiter Radius'), ('kilometer', 'kilometer'), ('m', 'meter'), ('R_earth', 'Earth Radius')], validators=[InputRequired('A stellar radius unit is required')])
    chemistry = SelectField('chemistry', choices=[('noTiO', '500'), ('eqchem', '750'), (1000, '1000'), (1250, '1250'), (1500, '1500'), (1750, '1750'), (2000, '2000'), (2250, '2250'), (2500, '2500')], validators=[InputRequired('A chemistry type is required')])
    clouds = SelectField('clouds', choices=[('0', 'Nothing'), ('ray10', 'Weak Rayleigh'), ('ray100', 'Medium Rayleigh'), ('ray1000', 'Strong Rayleigh'), ('flat10', 'Weak Cloud'), ('flat100', 'Medium Cloud'), ('flat1000', 'Strong Cloud')], validators=[InputRequired('A cloud model is required')])

    # Form submits
    calculate_submit = SubmitField('Calculate Forward Model')


class LimbDarkeningForm(BaseForm):
    """Form validation for the limb_darkening tool"""
    # Model grid
    modelgrid_dir = get_env_variables()['modelgrid_dir']
    default_modelgrid = os.path.join(modelgrid_dir, 'ATLAS9/')
    mg = ModelGrid(default_modelgrid, resolution=500)
    teff_rng = mg.Teff_vals.min(), mg.Teff_vals.max()
    logg_rng = mg.logg_vals.min(), mg.logg_vals.max()
    feh_rng = mg.FeH_vals.min(), mg.FeH_vals.max()
    modeldir = RadioField('modeldir', default=default_modelgrid, choices=[(os.path.join(modelgrid_dir, 'ATLAS9/'), 'Kurucz ATLAS9'), (os.path.join(modelgrid_dir, 'ACES/'), 'Phoenix ACES')], validators=[InputRequired('A model grid is required!')])

    # Stellar parameters
    teff = DecimalField('teff', default=3500, validators=[InputRequired('An effective temperature is required!'), NumberRange(min=float(teff_rng[0]), max=float(teff_rng[1]), message='Effective temperature must be between {} and {} for this model grid'.format(*teff_rng))])
    logg = DecimalField('logg', default=4.5, validators=[InputRequired('A surface gravity is required!'), NumberRange(min=float(logg_rng[0]), max=float(logg_rng[1]), message='Surface gravity must be between {} and {} for this model grid'.format(*logg_rng))])
    feh = DecimalField('feh', default=0.0, validators=[InputRequired('A surface gravity is required!'), NumberRange(min=float(feh_rng[0]), max=float(feh_rng[1]), message='Metallicity must be between {} and {} for this model grid'.format(*feh_rng))])
    mu_min = DecimalField('mu_min', default=0.1, validators=[InputRequired('A minimum mu value is required!'), NumberRange(min=0.0, max=1.0, message='Minimum mu must be between 0 and 1')])

    # LD profile
    profiles = MultiCheckboxField('profiles', choices=[(x, x) for x in PROFILES], validators=[InputRequired('At least one profile is required!')])

    # Bandpass
    default_filter = 'Kepler.K'
    defilt = svo.Filter(default_filter)
    bandpass = SelectField('bandpass', default=default_filter, choices=[('tophat', 'Top Hat')]+[(filt, filt) for filt in sorted(FILTERS['Band'])], validators=[InputRequired('A filter is required!')])
    wave_min = DecimalField('wave_min', default=defilt.wave_min.value, validators=[NumberRange(min=0, max=30, message='Minimum wavelength must be between 0 and 30 microns!')])
    wave_max = DecimalField('wave_max', default=defilt.wave_max.value, validators=[NumberRange(min=0, max=30, message='Maximum wavelength must be between 0 and 30 microns!')])
    n_bins = IntegerField('n_bins', default=1)

    # Form submits
    calculate_submit = SubmitField('Calculate Coefficients')
    filter_submit = SubmitField('Filter Selected')
    modelgrid_submit = SubmitField('Model Grid Selected')


class GroupsIntsForm(BaseForm):
    """Form validation for the groups_integrations tool"""
    # Form submits
    calculate_submit = SubmitField('Calculate Groups and Integrations')

    # Stellar Parameters
    kmag = DecimalField('kmag', default=10.5, validators=[InputRequired('A K-band magnitude is required!'), NumberRange(min=5.1, max=11.9, message='K-band mag must be between 5-12, non-inclusive.')])
    obs_duration = DecimalField('obs_duration', default=3, validators=[InputRequired('An observation duration is required!'), NumberRange(min=0, message='Observation duration must be a positive number')])
    time_unit = SelectField('time_unit', default='hour', choices=[('hour', 'hours'), ('day', 'days')])
    models = [('a0i', 'A0I 9750 2.0'), ('aov', 'A0V 9500 2.0'), ('a1v', 'A1V 9250 4.0'), ('a5i', 'A5I 8500 2.0'), ('a3v', 'A3V 8250 4.0'), ('a5v', 'A5V 8250 4.0'), ('f0i', 'F0I 7750 2.0'), ('f0v', 'F0V 7250 1.5'), ('f5i', 'F5I 7000 4.0'), ('f2v', 'F2V 7000 4.0'), ('f5v', 'F5V 6500 4.0'), ('f8v', 'F8V 6250 4.5'), ('g0v', 'G0V 6000 4.5'), ('g0iii', 'G0III 5750 3.0'), ('g2v', 'G2V 5750 4.5'), ('g5v', 'G5V 5750 4.5'), ('g0i', 'G0I 5500 1.5'), ('g8v', 'G8V 5500 4.5'), ('g5iii', 'G5III 5250 2.5'), ('g5i', 'G5I 4740 1.0'), ('k0v', 'K0V 5250 4.5'), ('k0iii', 'K0III 4750 2.0'), ('k2v', 'K2V 4750 4.5'), ('k0i', 'K0I 4500 1.0'), ('k5v', 'K5V 4250 1.5'), ('k5iii', 'K5III 4000 1.5'), ('k7v', 'K7V 4000 4.5'), ('k5i', 'K5I 3750 0.5'), ('m0i', 'M0I 3750 0.0'), ('m0iii', 'M0III 3750 1.5'), ('m0v', 'M0V 3750 4.5'), ('m2i', 'M2I 3500 0.0'), ('m2v', 'M2V 3500 4.5'), ('m5v', 'M5V 3500 5.0')]
    mod = SelectField('mod', choices=models)
    n_group = IntegerField('n_group', default=0)
    ins = SelectField('ins', default='miri', choices=[('niriss', 'NIRISS'), ('nircam', 'NIRCam'), ('nirspec', 'NIRSpec'), ('miri', 'MIRI')])

    # Filter selects
    miri_filt = SelectField('miri_filt', choices=[('lrs', 'LRS')])
    nirspec_filt = SelectField('nirspec_filt', choices=[('f070lp_g140h', 'F070LP/G140H'), ('f100lp_g140h', 'F100LP/G140H'), ('f070lp_g140m', 'F070LP/G140M'), ('f100lp_g140m', 'F100LP/G140M'), ('f170lp_g235h', 'F170LP/G235H'), ('f170lp_g235m', 'F170LP/G235M'), ('f290lp_g395h', 'F290LP/G395H'), ('f290lp_g395m', 'F290LP/G395M')])
    niriss_filt = SelectField('niriss_filt', choices=[('soss', 'SOSS')])
    nircam_filt = SelectField('nircam_filt', choices=[('f322w2', 'F322W2'), ('f444w', 'F444W'), ('f277w', 'F277W')])

    # TA filter selects
    miri_filt_ta = SelectField('miri_filt_ta', choices=[('f560w', 'F560W'), ('f100w', 'F100W'), ('f1500w', 'F1500W')])
    nirspec_filt_ta = SelectField('nirspec_filt_ta', choices=[('f110w', 'F110W'), ('f140x', 'F140X'), ('clear', 'CLEAR')])
    niriss_filt_ta = SelectField('niriss_filt_ta', choices=[('f480m', 'F480M')])
    nircam_filt_ta = SelectField('nircam_filt_ta', choices=[('f335m', 'F335M')])

    # Subarray selects
    miri_subarray = SelectField('miri_subarray', choices=[('slitlessprism', 'SLITLESSPRISM')])
    nirspec_subarray = SelectField('nirspec_subarray', choices=[('sub2048', 'SUB2048'), ('sub1024a', 'SUB1024A'), ('sub1024b', 'SUB1024B'), ('sub512', 'SUB512')])
    niriss_subarray = SelectField('niriss_subarray', choices=[('substrip256', 'SUBSTRIP256'), ('substrip96', 'SUBSTRIP96')])
    nircam_subarray = SelectField('nircam_subarray', choices=[('full', 'FULL FRAME'), ('subgrism256', 'SUBGRISM256'), ('subgrism128', 'SUBGRISM128'), ('subgrism64', 'SUBGRISM64')])

    # TA subarray selects
    miri_subarray_ta = SelectField('miri_subarray_ta', choices=[('slitlessprism', 'SLITLESSPRISM')])
    nirspec_subarray_ta = SelectField('nirspec_subarray_ta', choices=[('full', 'FULL'), ('sub32', 'SUB32'), ('sub2048', 'SUB2048')])
    niriss_subarray_ta = SelectField('niriss_subarray_ta', choices=[('nrm', 'SUBTASOSS -- BRIGHT'), ('im', 'SUBTASOSS -- FAINT')])
    nircam_subarray_ta = SelectField('nircam_subarray_ta', choices=[('sub32tats', 'SUB32TATS')])

    # Saturation
    sat_mode = RadioField('sat_mode', default='well', choices=[('counts', 'Counts'), ('well', 'Full well fraction')])
    sat_max = DecimalField('sat_max', default=0.95, validators=[InputRequired('A saturation level is required!'), NumberRange(min=0.0, message='Saturation level must be positive.')])


class ContamVisForm(BaseForm):
    """Form validation for the contamination_visibility tool"""
    # Form submits
    calculate_submit = SubmitField('Calculate Visibility')
    calculate_contam_submit = SubmitField('Calculate Visibility and Contamination')
    mode_submit = SubmitField('Mode Selected')

    # Form inputs
    ra = DecimalField('ra', validators=[NumberRange(min=0, max=360, message='RA must be between 0 and 360 degrees')])
    dec = DecimalField('dec', validators=[NumberRange(min=-90, max=90, message='Declinaton must be between -90 and 90 degrees')])
    inst = SelectField('inst', choices=[('NIRISS', 'NIRISS - SOSS'), ('NIRCam F322W2', 'NIRCam - Grism Time Series (F322W2)'), ('NIRCam F444W', 'NIRCam - Grism Time Series (F444W)'), ('MIRI', 'MIRI - LRS'), ('NIRSpec', 'NIRSpec (Visibility Only)')])
    companion = StringField('companion', default='')
    pa_min = DecimalField('pa_min', default=0, validators=[NumberRange(min=0, max=360, message='Minimum PA must be between 0 and 360 degrees')])
    pa_max = DecimalField('pa_max', default=360, validators=[NumberRange(min=0, max=360, message='Maximum PA must be between 0 and 360 degrees')])


class PhaseConstraint(BaseForm):

    calculate_submit = SubmitField('Calculate Phase Constraint')

    orbital_period = FloatField('orbital_period', validators=[InputRequired('Orbital period is a required field')]) 
    eccentricity = FloatField('eccentricity', default=np.nan)
    transit_type = SelectField('transit_type', choices=[('primary', 'primary'), ('secondary', 'secondary')])
    omega = FloatField('omega', default=np.nan)
    inclination = FloatField('inclination', default=np.nan)
    transit_time = FloatField('transit_time', default=np.nan)
    window_size = FloatField('window_size', default=1.0)
    observation_duration = FloatField('observation_duration', default=2.0, validators=[InputRequired('Observation duration is a required field.')])
    minimum_phase = DecimalField('minimum_phase', default=0.0)
    maximum_phase = DecimalField('maximum_phase', default=0.0)
    minimum_phase_sec = DecimalField('minimum_phase_sec', default=0.0)
    maximum_phase_sec = DecimalField('maximum_phase_sec', default=0.0)



"""
This is a module for calcuating groups and integrations with JWST on the fly. 
the main function (perform_calculation) takes a dictionary of inputs 
(modeled around how the web tool takes inputs) which must include : 
obs_time, n_group, mag, mod, band, filt, filt_ta, ins, subarray, subarray_ta, 
sat_mode, sat_max, and infile. It produces and dictionary of outputs that includes 
all of the original information as well as groups, integrations, saturation levels, 
and observation time estimates for target acquisition and science observations
with JWST. 


Authors
-------
    Jules Fowler, April 2017

Use
---
    This is mostly a module to be used by ExoCTKWeb -- but the main function
    can be run standalone.

"""

## -- IMPORTS
import json
import math
import os
from decimal import Decimal

from astropy.io import ascii
import numpy as np
from scipy import interpolate
from scipy.integrate import quad

## -- FUNCTIONS

def calc_groups_from_exp_time(max_exptime_per_int, t_frame):
    """
    Given the maximum saturation time, calculates the number
    of frames per group.

    Parameters
    ----------
    max_exptime_per_int : float
        The maximum number of seconds an integration can last
        before it's oversaturated. 
    t_frame : float
        The time per frame.

    Returns
    -------
    groups : int
        The required number of groups.
    """

    groups = max_exptime_per_int/t_frame
    return np.floor(groups)


def calc_n_int(transit_time, n_group, n_reset, t_frame, n_frame):
    """Calculates number of integrations required.

    Parameters
    ----------
    transit_time : float
        The time of your planeting transit (in hours.)
    n_group : int
        Groups per integration.
    n_reset : int
        Reset frames per integration.
    t_frame : float
        The frame time (in seconds).
    n_frame : int
        Frames per group -- always 1 except maybe brown dwarves.

    Returns
    -------
    n_ints : float
        The required number of integraitions.
    """

    hour = 3600
    n_ints = (float(transit_time)*hour)/(t_frame*(n_group*n_frame+n_reset))
    
    return math.ceil(n_ints)


def calc_obs_efficiency(t_exp, t_duration):
    """Calculates the observation efficiency.

    Parameters
    ----------
    t_exp : float
        Exposure time (in seconds).
    t_duration : float
        Duration time (in seconds).
    
    Returns
    -------
    obs_eff : float
        Observation efficiency.
    """

    obs_eff = t_exp/t_duration
    return obs_eff


def calc_t_duration(n_group, n_int, n_reset, t_frame, n_frame=1):
    """Calculates duration time (or exposure duration as told by APT.)

    Parameters
    ----------
    n_group : int
        Groups per integration.
    n_int : int
        Integrations per exposure.
    n_reset : int
        Reset frames per integration.
    t_frame : float
        Frame time (in seconds).
    n_frame : int, optional
        Frames per group -- always one except brown dwarves.

    Returns
    -------
    t_duration : float
        Duration time (in seconds).
    """
    
    t_duration = t_frame*(n_group*n_frame + n_reset)*n_int
    return t_duration


def calc_t_exp(n_int, t_ramp):
    """Calculates exposure time (or photon collection duration as told by APT.)

    Parameters
    ----------
    n_int : int
        Integrations per exposure.
    t_ramp : float
        Ramp time (in seconds).

    Returns
    -------
    t_exp : float
        Exposure time (in seconds).
    """

    t_exp = n_int*t_ramp
    return t_exp


def calc_t_frame(n_col, n_row, n_amp, ins):
    """Calculates the frame time for a given ins/readmode/subarray.

    Parameters
    ----------
    n_col : int
        Number of columns.
    n_row : int
        Number of rows.
    n_amp : int
        Amplifiers reading data.
    ins : str
        The instrument key.

    Returns:
    t_frame : float
        The frame time (in seconds).
    """
    n_col, n_amp, n_row = int(n_col), int(n_amp), int(n_row)
    
    if ins == 'nirspec':
        n = 2
    if ins in ['nircam', 'niriss']:
        n = 1

    t_frame = (n_col/n_amp + 12)*(n_row + n)*(1e-5)
    
    return t_frame


def calc_t_int(n_group, t_frame, n_frame, n_skip):
    """Calculates the integration time.
    
    Parameters
    ----------
    n_group : int
        Groups per integration.
    t_frame : float
        Frame time (in seconds).
    n_frame : int
        Frames per group -- always 1 except maybe brown dwarves.
    n_skip : int 
        Skips per integration -- always 0 except maybe brown dwarves.

    Returns
    -------
    t_int : float
        Integration time (in seconds). 
    """
    
    t_int = (n_group*(n_frame + n_skip) - n_skip)*t_frame
    return t_int


def calc_t_ramp(t_int, n_reset, t_frame):
    """Calculates the ramp time -- or the integration time plus overhead for resets.

    Parameters
    ----------
    t_int : float
        Integration time (in seconds).
    n_reset : int
        Rest frames per integration
    t_frame : float
        Frame time (in seconds).

    Returns
    -------
    t_ramp : float
        Ramp time (in seconds).
    """

    t_ramp = t_int + (n_reset - 1)*t_frame
    return t_ramp


def convert_sat(sat_max, sat_mode, ins, infile, ta=False):
    """ Converts full well fraction to a saturation in counts OR provides the 
    max fullwell for TA mode.

    Parameters
    ----------
    sat_max : float
        Either a full well fraction or counts.
    sat_mode : str
        'well' or 'counts'.
    ins : str
        The instrument.
    infile : str
        The path to the data file.
    ta : bool, optional
        Whether or not it's TA mode.
    
    Returns
    -------
    sat_max : float
        The fullwell to use in counts.
    """

    with open(infile) as f:
        dat = json.load(f)

    ins_dict = dat['fullwell'] 
    
    if sat_mode == 'well':
        sat_max = float(sat_max)*float(ins_dict[ins])
    
    if ta:
        sat_max = ins_dict[ins]
    
    return sat_max
    

def interpolate_from_dat(mag, ins, filt, sub, mod, band, t_frame, sat_lvl, infile, ta=False):
    """
    Interpolates the precalculated pandeia data to estimate the saturation limit.

    Parameters
    ----------
    mag : float
        The magnitude of the source. (Takes between 4.5-12.5)
    ins : str
       The instrument, allowable "miri", "niriss", "nirspec", "nircam". 
    filt : str
        Filter.
    sub : str   
        Subarray.
    mod : str
        Phoenix model key.
    band : str
        Magnitude band -- unsed rn.
    t_frame : float
        Frame time.
    sat_lvl : float 
        The maximum fullwell saturation we'll allow.
    infile : str
        The data file to use.
    ta : bool, optional
        Whether or not we're running this for TA.

    Returns
    -------
    n_group : int
        The number of groups that won't oversaturate the detector.
    max_sat : int
        The maximum saturation level reached by that number of groups.
    """
    # Create the dictionaries for each filter and select out the prerun data
    with open(infile) as f:
        dat = json.load(f)

    ta_or_sci = 'sci_sat'

    if ta:
        ta_or_sci = 'ta_sat'

    # The data
    mags = np.array(dat['mags'])
    sat = dat[ta_or_sci][ins][filt][sub][mod]
    log_sat = np.log10(sat)

    # Interpolate the given magnitude
    func_log = interpolate.interp1d(mags, log_sat)
    max_log_sat = func_log(float(mag))
    max_sat = 10**(max_log_sat)

    # Figure out what it means in wake of the given sat lvl
    max_exptime = float(sat_lvl)/max_sat

    # Calculate the nearest number of groups
    n_group = calc_groups_from_exp_time(max_exptime, t_frame)

    # Can't have zero groups
    n_group = n_group or 1

    return n_group, max_sat


def map_to_ta_modes(ins, max_group, min_group):
    """Turns the min/max groups into the closest allowable
    TA group mode.

    Parameters
    ----------
    ins : str
        Instrument.
    max_group : int
        The maximum number of groups without oversaturating.
    
    min_group : int
        The groups needed to hit the target SNR.
    
    Returns
    -------
    min_ta_groups : int
        The min possible groups to hit target SNR.
    max_ta_groups : int
        The max possible groups before saturation.
    """

    # Allowable group modes for each ins
    groups = {'miri': [3, 5, 9, 15, 23, 33, 45, 59, 75, 93, 113, 135, 159, 185, 243, 275, 513],
              'niriss': [3, 5, 7, 9, 1, 13, 15, 17, 19],
              'nirspec': [3], 
              'nircam': [3, 5, 9, 17, 33, 65]
              }
    
    # Match the literal min and max groups to the nearest mode. 
    allowable_groups = groups[ins]
    min_ta_groups = min(allowable_groups, key=lambda x:abs(x-min_group))
    max_ta_groups = min(allowable_groups, key=lambda x:abs(x-max_group))
    
    # Unless it was oversaturated from the get-go OR there aren't enough groups
    # for SNR
    if min_group == 0:
        min_ta_groups = 0
        max_ta_groups = 0
    if min_group > max(allowable_groups):
        min_ta_groups = -1
        max_ta_groups = 0
    
    # BOTH ARE FLIPPED RN -- I WILL FLIP BOTH BACK SOON...
    return max_ta_groups, min_ta_groups


def min_groups(mag, ins, filt, sub, mod, band, infile):
    """Estimates the minimum number of groups to reach 
    target acq sat requirements.

    Parameters
    ----------
    mag : float 
        Magnitude of star.
    ins : str
        Instrument.
    filt : str
        Filter.
    sub : str
        Subarray.
    mod : str
        Phoenix model key.
    band : str, currently unused?
        The band -- right now only k sooo?
    infile : str
        The file with the pandeia data.
    
    Returns
    -------
    min_groups : int
        The minimum number of groups to reach target snr.
    """

    with open(infile) as f:
        dat = json.load(f)

    # Match to closest magnitude
    mags = [float(i) for i in dat['mags']]
    closest_mag = min(mags, key=lambda x:abs(x-float(mag)))
    index = mags.index(closest_mag)

    # Match to data 
    min_groups = dat['ta_snr'][ins][filt][sub][mod][index]

    return min_groups


def perform_calculation(params, n_frame=1, n_skip=0):
    """Calculates all of the outputs and puts them in a dictionary for easy access. 

    Parameters
    ----------
    params : dict
        Dictionary of all the needed parameters. Must include:
        obs_time, n_group, mag, mod, band, filt, filt_ta,
        ins, subarray, subarray_ta, sat_mode, sat_max, infile
    n_frame :int, optional
        The number of frames -- almost always 1.
    n_skip: int, optional
        Number of skips -- almost always 0
    
    Returns
    -------
    params : dict, str
        Dictionary of outputs and inputs. If 
        the calculation throws an error it will return a string 
        error message instead.
    """
    
    ## -- TARGET ACQ
    ta_frame_time = set_t_frame(params['infile'], params['ins'], params['subarray_ta'], ta=True)
    
    max_sat_ta_lvl = convert_sat(params['sat_max'], params['sat_mode'],
            params['ins'], params['infile'], ta=True)

    max_group, sat_rate_ta = interpolate_from_dat(params['mag'], params['ins'], params['filt_ta'], 
                                     params['subarray_ta'], params['mod'], params['band'],
                                     ta_frame_time, max_sat_ta_lvl, params['infile'], ta=True)
    min_group = min_groups(params['mag'], params['ins'], params['filt_ta'], params['subarray_ta'], 
                           params['mod'], params['band'], params['infile'])
    min_ta_groups, max_ta_groups = map_to_ta_modes(params['ins'], max_group, min_group)
    t_duration_ta_min = calc_t_duration(min_ta_groups, 1, 1, ta_frame_time)
    t_duration_ta_max = calc_t_duration(max_ta_groups, 1, 1, ta_frame_time)

    ## -- Science obs
    
    # Figure out the rows/cols/amps/px_size and sat
    ins_params = set_params_from_ins(params['ins'], params['subarray'])
    frame_time = set_t_frame(params['infile'], params['ins'], params['subarray'])

    # Run all the calculations
    n_row, n_col, n_amp, px_size, t_frame, n_reset = ins_params
    band_ins = '{0}_{1}'.format(params['band'], params['filt'])
    
    params['sat_max'] = convert_sat(params['sat_max'], params['sat_mode'], params['ins'], params['infile'])

    # Calculate countrate and n_groups if it isn't supplied
    n_group, sat_rate = interpolate_from_dat(params['mag'],
        params['ins'], params['filt'], params['subarray'], params['mod'],
        params['band'], frame_time, params['sat_max'], params['infile'])

    if str(params['n_group']) == 'optimize':
        params['n_group'] = int(n_group)
    else:
        params['n_group'] = int(float(params['n_group']))
    
    ## -- Aditional helpful params

    # Calculate times/ramps/etc
    t_int = calc_t_int(params['n_group'], frame_time, n_frame, n_skip)
    t_ramp = calc_t_ramp(t_int, n_reset, frame_time)
    
    # Calculate nubmer of integrations (THE MEAT)
    n_int = calc_n_int(params['obs_time'], params['n_group'], n_reset, frame_time, n_frame)
    
    # Other things that may come in handy who knows?
    t_exp = calc_t_exp(n_int, t_ramp)
    t_duration = calc_t_duration(params['n_group'], n_int, n_reset, frame_time, n_frame)
    obs_eff = calc_obs_efficiency(t_exp, t_duration)

    # Update params with new friends
    params['n_col'] = n_col
    params['n_row'] = n_row
    params['n_amp'] = n_amp
    params['n_reset'] = n_reset
    params['n_frame'] = n_frame
    params['n_skip'] = n_skip
    params['t_frame'] = round(frame_time, 3)
    params['t_int'] = round(t_int, 3)
    params['t_ramp'] = round(t_ramp, 3)
    params['n_int'] = n_int
    params['t_exp'] = round(t_exp/3600, 3)
    params['t_duration'] = round(t_duration/3600, 3)
    params['obs_eff'] = round(obs_eff, 3)
    params['ta_t_frame'] = ta_frame_time
    params['min_ta_groups'] = int(min_ta_groups)
    params['max_ta_groups'] = int(max_ta_groups)
    params['t_duration_ta_min'] = t_duration_ta_min
    params['t_duration_ta_max'] = t_duration_ta_max
    params['max_sat_prediction'] = round(sat_rate*frame_time*params['n_group'], 3)
    params['max_sat_ta'] = round(sat_rate_ta*ta_frame_time*max_ta_groups, 3)
    params['min_sat_ta'] = round(sat_rate_ta*ta_frame_time*min_ta_groups, 3)
    
    return params 


def set_params_from_ins(ins, subarray):
    """Sets/collects the running parameters from the instrument.
    
    Parameters
    ----------
    ins : str
        Instrument, options are nircam, niriss, nirpec, and miri.
    subarray : str
        Subarray mode.

    Returns
    -------
    rows : int  
        The number of pixels per row.
    cols : int
        The number of columns per row.
    """
    
    n_reset = 1
 
    if ins == 'nirspec':
        px_size = (40e-4)**2

        if subarray == 'sub2048':
            rows, cols = 2048, 32
        elif subarray in ['sub1024a', 'sub1024b']:
            rows, cols = 1024, 32
        elif subarray == 'sub512':
            rows, cols = 512, 32
        
        amps = 1 # 4 if not NRSRAPID????
        ft = calc_t_frame(cols, rows, amps, ins)
  
    elif ins == 'nircam':
        px_size = (18e-4)**2
        
        if subarray == 'full':
            rows, cols, amps = 2048, 2048, 4
        elif subarray == 'subgrism256':
            rows, cols, amps = 256, 256, 1
        elif subarray == 'subgrism128':
            rows, cols, amps = 128, 2048, 1
        elif subarray == 'subgrism64':
            rows, cols, amps = 64, 2048, 1
        
        ft = calc_t_frame(cols, rows, amps, ins)
  
    elif ins == 'miri':
        px_size = (25e-4)**2
        
        if subarray == 'slitlessprism':
            rows, cols, ft = 416, 72, .159
        
        amps = 4
        n_reset = 0 
        
    elif ins == 'niriss':
        px_size = (40e-4)**2
        
        if subarray == 'substrip96':
            rows, cols = 2048, 96
        elif subarray == 'substrip256':
            rows, cols = 2048, 256
        
        amps = 1
        ft = calc_t_frame(cols, rows, amps, ins)
    
    return rows, cols, amps, px_size, ft, n_reset
    

def set_t_frame(infile, ins, sub, ta=False):
    """ Assign the appropriate frame time based on the ins
    and subarray. For now, modes are implied.

    Parameters
    ----------
    infile: str
        The path to the data file.
    ins : str
        The instrument : 'miri', 'niriss', 'nirspec', or 
        'nircam'.
    sub : str
        The subarray -- too lazy to write out the options 
        here.
    ta : bool   
        Whether this is for TA or not.

    Returns
    -------
    t_frame : float
        The frame time for this ins/sub combo.
    """
    
    # Read in dict with frame times
    with open(infile) as f:
        frame_time = json.load(f)['frame_time']

    if ta:
        t_frame = frame_time[ins]['ta'][sub]
    else:
        t_frame = frame_time[ins][sub]

    return t_frame


## -- RUN

# There's no function call right now because these inputs are ornery. 
# I can add one if someone feels strongly.



#! /usr/bin/env python

"""This module provides examples of how to use the ``platon_wrapper``
module in ExoCTK's ``atmopsheric_retrieval`` subpackage.  Each example
can be run using either the ``multinest`` method or the ``emcee``
method.  See the docstring of each example function for further
details.

Authors
-------

    - Matthew Bourque

Use
---

    To run all examples, one can execute this script via the command
    line as such:

        >>> python examples.py

    To run individual examples, one can import the example functions
    and pass as a parameter which method to run.  Available
    examples include:

        from examples import example, example_aws_short, example_aws_long
        example('emcee')
        example('multinest')
        example_aws_short('emcee')
        example_aws_short('multinest')
        example_aws_long('emcee')
        example_aws_long('multinest')

Dependencies
------------

    Dependent libraries include:

    - ``exoctk``
    - ``numpy``
    - ``pandas``
    - ``platon``

    To run the examples that use AWS, users must also have a
    ``aws_config.json`` file present within the
    ``atmospheric_retrievals`` subdirectory.  This file must be of
    a valid JSON format and contain two key/value pairs,
    ``ec2_id`` and ``ssh_file``, e.g.:

    {
    "ec2_id" : "lt-021de8b904bc2b728",
    "ssh_file" : "~/.ssh/my_ssh_key.pem"
    }

    where the ``ec2_id`` contains the ID for an EC2 launch template
    or an existing EC2 instance, and ``ssh_file`` points to the SSH
    public key used for logging into an AWS account.

    Note that if the ``ec2_id`` points to a launch template (i.e. the
    string starts with ``lt-``), a new EC2 instance will be created and
    launched.  However, if the ``ec2_id`` points to an existing EC2
    instance (i.e. the string starts with ``i-``), the existing EC2
    instance will be started and used.

References
----------

    Example data was pulled from the "Transmission Spectra" tab from
    corresponding ExoMAST pages, available at
    https://exo.mast.stsci.edu/
"""

import logging
import os
import time

import numpy as np
import pandas
from platon.constants import R_sun, R_jup, M_jup

from exoctk.atmospheric_retrievals.aws_tools import get_config
from exoctk.atmospheric_retrievals.platon_wrapper import PlatonWrapper


def example(method):
    """Performs a short example run of the retrievals using local
    machine.

    Parameters
    ----------
    method : str
        The method to use to perform the atmopsheric retrieval; can
        either be ``multinest`` or ``emcee``
    """

    # Ensure that the method parameter is valid
    assert method in ['multinest', 'emcee'], \
        'Unrecognized method: {}'.format(method)

    # Define the fit parameters
    params = {
        'Rs': 1.19,  # Required
        'Mp': 0.73,  # Required
        'Rp': 1.4,  # Required
        'T': 1200.0,  # Required
        'logZ': 0,  # Optional
        'CO_ratio': 0.53,  # Optional
        'log_cloudtop_P': 4,  # Optional
        'log_scatt_factor': 0,  # Optional
        'scatt_slope': 4,  # Optional
        'error_multiple': 1,  # Optional
        'T_star': 6091}  # Optional

    # Initialize the object and set the parameters
    pw = PlatonWrapper()
    pw.set_parameters(params)

    # Fit for the stellar radius and planetary mass using Gaussian priors.  This
    # is a way to account for the uncertainties in the published values
    pw.fit_info.add_gaussian_fit_param('Rs', 0.02*R_sun)
    pw.fit_info.add_gaussian_fit_param('Mp', 0.04*M_jup)

    # Fit for other parameters using uniform priors
    pw.fit_info.add_uniform_fit_param('Rp', 0.9*(1.4 * R_jup), 1.1*(1.4 * R_jup))
    pw.fit_info.add_uniform_fit_param('T', 0.5*1200, 1.5*1200)
    pw.fit_info.add_uniform_fit_param("log_scatt_factor", 0, 1)
    pw.fit_info.add_uniform_fit_param("logZ", -1, 3)
    pw.fit_info.add_uniform_fit_param("log_cloudtop_P", -0.99, 5)
    pw.fit_info.add_uniform_fit_param("error_multiple", 0.5, 5)

    # Define bins, depths, and errors
    pw.wavelengths = 1e-6*np.array([1.119, 1.1387])
    pw.bins = [[w-0.0095e-6, w+0.0095e-6] for w in pw.wavelengths]
    pw.depths = 1e-6 * np.array([14512.7, 14546.5])
    pw.errors = 1e-6 * np.array([50.6, 35.5])

    # Do some retrievals
    pw.retrieve(method)

    # Save the results
    pw.save_results()

    # Make corner plot of results
    pw.make_plot()

    return pw


def example_aws_short(method):
    """Performs an short example run of the retrievals using AWS.

    Parameters
    ----------
    method : str
        The method to use to perform the atmopsheric retrieval; can
        either be ``multinest`` or ``emcee``
    """

    # Define the fit parameters
    params = {
        'Rs': 1.19,  # Required
        'Mp': 0.73,  # Required
        'Rp': 1.4,  # Required
        'T': 1200.0,  # Required
        'logZ': 0,  # Optional
        'CO_ratio': 0.53,  # Optional
        'log_cloudtop_P': 4,  # Optional
        'log_scatt_factor': 0,  # Optional
        'scatt_slope': 4,  # Optional
        'error_multiple': 1,  # Optional
        'T_star': 6091}  # Optional

    # Initialize the object, set parameters, and perform retrieval
    pw = PlatonWrapper()
    pw.set_parameters(params)

    # Fit for the stellar radius and planetary mass using Gaussian priors.  This
    # is a way to account for the uncertainties in the published values
    pw.fit_info.add_gaussian_fit_param('Rs', 0.02*R_sun)
    pw.fit_info.add_gaussian_fit_param('Mp', 0.04*M_jup)

    # Fit for other parameters using uniform priors
    pw.fit_info.add_uniform_fit_param('Rp', 0.9*(1.4 * R_jup), 1.1*(1.4 * R_jup))
    pw.fit_info.add_uniform_fit_param('T', 0.5*1200, 1.5*1200)
    pw.fit_info.add_uniform_fit_param("log_scatt_factor", 0, 1)
    pw.fit_info.add_uniform_fit_param("logZ", -1, 3)
    pw.fit_info.add_uniform_fit_param("log_cloudtop_P", -0.99, 5)
    pw.fit_info.add_uniform_fit_param("error_multiple", 0.5, 5)

    # Define bins, depths, and errors
    pw.wavelengths = 1e-6*np.array([1.119, 1.1387])
    pw.bins = [[w-0.0095e-6, w+0.0095e-6] for w in pw.wavelengths]
    pw.depths = 1e-6 * np.array([14512.7, 14546.5])
    pw.errors = 1e-6 * np.array([50.6, 35.5])

    # Set use for AWS and perform retrieval
    ssh_file = get_config()['ssh_file']
    ec2_id = get_config()['ec2_id']
    pw.use_aws(ssh_file, ec2_id)
    pw.retrieve(method)


def example_aws_long(method):
    """Performs an longer example run of the retrievals using AWS.

    Parameters
    ----------
    method : str
        The method to use to perform the atmopsheric retrieval; can
        either be ``multinest`` or ``emcee``
    """

    params = {
        'Rs': 1.19,
        'Mp': 0.73,
        'Rp': 1.39,
        'T': 1476.81,
        'logZ': 0,
        'CO_ratio': 0.53,
        'log_cloudtop_P': 4,
        'log_scatt_factor': 0,
        'scatt_slope': 4,
        'error_multiple': 1,
        'log_cloudtop_P': 4}

    # Initialize the object, set parameters, and perform retrieval
    pw = PlatonWrapper()
    pw.set_parameters(params)

    if method == 'multinest':
        pw.fit_info.add_gaussian_fit_param('Rs', 0.02*R_sun)
        pw.fit_info.add_gaussian_fit_param('Mp', 0.04*M_jup)
        pw.fit_info.add_uniform_fit_param('Rp', 0.9*(1.39 * R_jup), 1.1*(1.39 * R_jup))
        pw.fit_info.add_uniform_fit_param('T', 300, 3000)
        pw.fit_info.add_uniform_fit_param("log_scatt_factor", 0, 2)
        pw.fit_info.add_uniform_fit_param("logZ", -1, 3)
        pw.fit_info.add_uniform_fit_param("log_cloudtop_P", -0.99, 7)
        pw.fit_info.add_uniform_fit_param("error_multiple", 0.5, 5)
    elif method == 'emcee':
        pw.fit_info.add_gaussian_fit_param('Rs', 0.02*R_sun)
        pw.fit_info.add_gaussian_fit_param('Mp', 0.04*M_jup)
        pw.fit_info.add_uniform_fit_param('Rp', 0, np.inf, 0.9*(1.39 * R_jup), 1.1*(1.39 * R_jup))
        pw.fit_info.add_uniform_fit_param('T', 300, 3000, 0.5*1476.81, 1.5*1476.81)
        pw.fit_info.add_uniform_fit_param("log_scatt_factor", 0, 5, 0, 2)
        pw.fit_info.add_uniform_fit_param("logZ", -1, 3)
        pw.fit_info.add_uniform_fit_param("log_cloudtop_P", -0.99, 7)
        pw.fit_info.add_uniform_fit_param("error_multiple", 0, np.inf, 0.5, 5)

    # Get bins, depths, and errors
    bins, depths, errors = get_example_data('hd209458b')
    pw.bins = bins
    pw.depths = depths
    pw.errors = errors

    # Set use for AWS and perform retrieval
    ssh_file = get_config()['ssh_file']
    ec2_id = get_config()['ec2_id']
    pw.use_aws(ssh_file, ec2_id)
    pw.retrieve(method)


def get_example_data(object_name):
    """Return ``bins``, ``depths``, and ``errors`` for the given
    ``object_name``.  Data is read in from a ``csv`` file with a
    filename corresponding to ``object_name``.

    Parameters
    ----------
    object_name : str
        The object of interest (e.g. ``hd209458b``)

    Returns
    -------
    bins : np.array
        A 2xN ``numpy`` array of wavelength bins, of the form
        ``[[wavelength_bin_min, wavelength_bin_max], ...]``
    depths : np.array
        A 1D ``numpy`` array of depth values
    errors: np.array
        A 1D ``numpy`` array of depth error values.
    """

    logging.info('Using data for {}'.format(object_name))

    # Read in the data
    data_file = os.path.join(os.path.dirname(__file__), 'example_data', '{}.csv'.format(object_name))
    df = pandas.read_csv(data_file, names=['wavelengths', 'bin_sizes', 'depths', 'errors'])

    # Remove and rows outside of wavelength range (3e-7 to 3e-5)
    df = df.loc[(1e-6*df['wavelengths'] - 1e-6*df['bin_sizes'] >= 3e-7) & (1e-6*df['wavelengths'] + 1e-6*df['bin_sizes'] <= 3e-5)]

    # Parse the data
    wavelengths = 1e-6*np.array(df['wavelengths'])
    bin_sizes = 1e-6*np.array(df['bin_sizes'])
    depths = np.array(df['depths'])
    errors = np.array(df['errors'])

    # Calculate bins
    bins = [[wavelength - bin_size, wavelength + bin_size] for wavelength, bin_size in zip(wavelengths, bin_sizes)]

    return bins, depths, errors


if __name__ == '__main__':

    # A short example using local machine
    example('multinest')
    example('emcee')

    # A short example using AWS
    example_aws_short('multinest')
    time.sleep(120)  # Allow time for the EC2 instance to restart
    example_aws_short('emcee')

    # A long example using AWS
    time.sleep(120)  # Allow time for the EC2 instance to restart
    example_aws_long('multinest')
    time.sleep(120)  # Allow time for the EC2 instance to restart
    example_aws_long('emcee')



from distutils.extension import Extension

def get_package_data():
    return {'exoctk': ['data/*', 'data/images/*', 'data/core/*', 'data/core/modelgrid/*', 'data/contam_visibility/*', 'exoctk_app/static/images/*', 'exoctk_app/static/css/*', 'exoctk_app/static/js/*', 'exoctk_app/static/js/vendor/*']}
